"""
colours.py - Colour definitions for the Python ldraw package.

Copyright (C) 2008 David Boddie <david@boddie.org.uk>

This file is part of the ldraw Python package.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


class Colour(object):
    # pylint: disable=too-many-arguments, too-few-public-methods
    """ a Colour, uniquely identified by a code """
    def __init__(self, code=None, name=None, rgb=None, alpha=None, colour_attributes=None):
        self.code = code
        self.name = name
        self.rgb = rgb
        self.alpha = alpha
        self.colour_attributes = colour_attributes

    def __eq__(self, other):
        if isinstance(other, Colour):
            return self.code == other.code
        return self.code == other

    def __hash__(self):
        return hash(self.code)



"""
figure.py - Mini-figure construction classes for the ldraw Python package.

Copyright (C) 2008 David Boddie <david@boddie.org.uk>

This file is part of the ldraw Python package.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
#pylint: disable=missing-docstring
from ldraw.geometry import Vector, Identity, XAxis, YAxis, ZAxis
from ldraw.library.parts.minifig.accessories import Airtanks, HipsAndLegs, Hips
from ldraw.library.parts.minifig.arms import ArmLeft, ArmRight
from ldraw.library.parts.minifig.hands import Hand
from ldraw.library.parts.minifig.legs import LegLeft, LegRight
from ldraw.library.parts.minifig.torsos import Torso
from ldraw.pieces import Piece
from ldraw.library.parts.minifig.heads import HeadWithSwSmirkAndBrownEyebrowsPattern as Head


def dependent_piece(dep):
    def decorator(fn):
        def wrapped(self, *args, **kwargs):
            try:
                dependent_object = self.pieces_info[dep]
                return fn(self, dependent_object, *args, **kwargs)
            except KeyError:
                return None
        return wrapped
    return decorator


class Person(object):
    def __init__(self, position=Vector(0, 0, 0), matrix=Identity(),
                 group=None):
        self.position = position
        self.matrix = matrix
        self.pieces_info = {}
        self.group = group

    def head(self, colour, angle=0, part=Head):
        """ Displacement from torso """
        displacement = self.matrix * Vector(0, -24, 0)
        piece = Piece(colour, self.position + displacement,
                      self.matrix * Identity().rotate(angle, YAxis), part,
                      self.group)
        self.pieces_info["head"] = piece
        return piece

    @dependent_piece('head')
    def hat(self, head, colour, part="3901"):
        # Displacement from head
        displacement = head.position + head.matrix * Vector(0, 0, 0)
        piece = Piece(colour, displacement, head.matrix, part, self.group)
        return piece

    def torso(self, colour, part=Torso):
        """ torso piece """
        return Piece(colour, self.position, self.matrix, part, self.group)

    def backpack(self, colour, displacement=Vector(0, 0, 0), part=Airtanks):
        """ Displacement from torso """
        displacement = self.matrix * displacement
        return Piece(colour, self.position + displacement, self.matrix, part,
                     self.group)

    def hips_and_legs(self, colour, part=HipsAndLegs):
        """ Displacement from torso """
        displacement = self.matrix * Vector(0, 32, 0)
        return Piece(colour, self.position + displacement, self.matrix, part,
                     self.group)

    def hips(self, colour, part=Hips):
        """ Displacement from torso """
        displacement = self.matrix * Vector(0, 32, 0)
        return Piece(colour, self.position + displacement, self.matrix, part,
                     self.group)

    def left_arm(self, colour, angle=0, part=ArmLeft):
        """ Displacement from torso """
        displacement = self.matrix * Vector(15, 8, 0)
        piece = Piece(colour, self.position + displacement,
                      self.matrix * Identity().rotate(-10, ZAxis) *
                      Identity().rotate(angle, XAxis), part, self.group)
        self.pieces_info["left arm"] = piece
        return piece

    @dependent_piece('left arm')
    def left_hand(self, left_arm, colour, angle=0, part=Hand):
        # Displacement from left hand
        displacement = left_arm.position + left_arm.matrix * Vector(4, 17, -9)
        matrix = left_arm.matrix * Identity().rotate(40, XAxis) * Identity().rotate(angle, ZAxis)
        piece = Piece(colour, displacement, matrix, part, self.group)
        self.pieces_info["left hand"] = piece
        return piece

    @dependent_piece('left hand')
    def left_hand_item(self, left_hand, colour, displacement, angle=0, part=None):
        """ Displacement from left hand """
        if not part:
            return None
        # Displacement from left hand
        displacement = left_hand.position + left_hand.matrix * displacement
        matrix = left_hand.matrix * Identity().rotate(10, XAxis) * Identity().rotate(angle, YAxis)
        piece = Piece(colour, displacement, matrix, part, self.group)
        return piece

    def right_arm(self, colour, angle=0, part=ArmRight):
        """ Displacement from torso """
        displacement = self.matrix * Vector(-15, 8, 0)
        piece = Piece(colour, self.position + displacement,
                      self.matrix * Identity().rotate(10, ZAxis) *
                      Identity().rotate(angle, XAxis), part, self.group)
        self.pieces_info["right arm"] = piece
        return piece

    @dependent_piece('right arm')
    def right_hand(self, right_arm, colour, angle=0, part=Hand):
        # Displacement from right arm
        displacement = right_arm.position + right_arm.matrix * Vector(-4, 17, -9)
        matrix = right_arm.matrix * Identity().rotate(40, XAxis) * Identity().rotate(angle, ZAxis)
        piece = Piece(colour, displacement, matrix, part, self.group)
        self.pieces_info["right hand"] = piece
        return piece

    @dependent_piece('right hand')
    def right_hand_item(self, right_hand, colour, displacement, angle=0, part=None):
        """ Add a right hand item"""
        if not part:
            return None
        # Displacement from right hand
        displacement = right_hand.position + right_hand.matrix * displacement
        matrix = right_hand.matrix * Identity().rotate(10, XAxis) * Identity().rotate(angle, YAxis)
        piece = Piece(colour, displacement, matrix, part, self.group)
        return piece

    def left_leg(self, colour, angle=0, part=LegLeft):
        """ Add a left leg """
        displacement = self.matrix * Vector(0, 44, 0)
        piece = Piece(colour, self.position + displacement,
                      self.matrix * Identity()
                      .rotate(angle, XAxis), part,
                      self.group)
        self.pieces_info["left leg"] = piece
        return piece

    @dependent_piece('left leg')
    def left_shoe(self, left_leg, colour, angle=0, part=None):
        """ Add a shoe on the left"""
        if not part:
            return None
        # Displacement from left leg
        displacement = left_leg.position + left_leg.matrix * Vector(10, 28, 0)
        matrix = left_leg.matrix * Identity().rotate(angle, YAxis)
        piece = Piece(colour, displacement, matrix, part, self.group)
        return piece

    def right_leg(self, colour, angle=0, part=LegRight):
        """" Add a right leg"""
        displacement = self.matrix * Vector(0, 44, 0)
        piece = Piece(colour, self.position + displacement,
                      self.matrix * Identity()
                      .rotate(angle, XAxis), part,
                      self.group)
        self.pieces_info["right leg"] = piece
        return piece

    @dependent_piece('right leg')
    def right_shoe(self, right_leg, colour, angle=0, part=None):
        """ Add a shoe on the right """
        if not part:
            return None
        # Displacement from right leg
        displacement = right_leg.position + right_leg.matrix * Vector(-10, 28, 0)
        matrix = right_leg.matrix * Identity().rotate(angle, YAxis)
        piece = Piece(colour, displacement, matrix, part, self.group)
        return piece



# coding=utf-8
"""
library/parts.py - Auto-generated Part classes for the Python ldraw package.

Copyright (C) 2018 Matthieu Berthomé <rienafairefr@gmail.com>

This file is part of the ldraw Python package.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


<< This file is auto-generated, you shouldn't have to modify it>>
"""
# pylint: disable=invalid-name, line-too-long, too-many-lines

MechanicalCuboidTorsoWithStudOnFront = "24078"
MechanicalHeadAndTorso = "53988"
MechanicalTorso = "30375"
MechanicalTorsoWith4SideAttachmentCylinders = "54275"
MechanicalTorsoWithBlueInsigniaPattern = "30375ps3"
MechanicalTorsoWithOrangeDotInsigniaPattern = "30375ps4"
MechanicalTorsoWithOrangeInsigniaPattern = "30375ps1"
MechanicalTorsoWithRedDotInsigniaPattern = "30375ps7"
MechanicalTorsoWithTanInsigniaPattern = "30375ps2"
MechanicalTorsoWithTanThreeDotInsigniaPattern = "30375ps6"
MechanicalTorsoWithYellowDotInsigniaPattern = "30375ps5"
SkeletonTorso = "6260"
SkeletonTorsoWithShoulderRods = "60115"
Torso = "973"
TorsoBatmanLogoInYellowOvalWithYellowUtilityBeltAndGoldBucklePattern = "973pbo"
TorsoFemaleBlouseWithButtons_DarkPinkPanelAndNameBadgePattern = "973p9a"
TorsoFemaleTopWith2WhiteButtonsAndBlackStrapsPattern = "973p9j"
TorsoFemaleTopWithDarkPinkStripesAndFlowerNecklacePattern = "973p1z"
TorsoFemaleWithGoldStarAndZipperAndRedTiedSashAndMusclesOutlinePattern = "973pc1"
TorsoFemaleWithHarlequinBlack_WhitePattern = "973pba"
TorsoFemaleWithJacket_2Pockets_Buttons_GoldNecklacePattern = "973pbj"
TorsoFemaleWithLabCoatOverMediumBlueShirtWithIdBadgeAndSpaceShuttleLogoOnBackPattern = "973p8q"
TorsoFemaleWithLtpurpleScarfAndGrayBeltPattern = "973p7y"
TorsoFemaleWithTopWithYellowNeckAndWhiteUndershirtPattern = "973p9f"
TorsoHpJacketWith4ButtonVestAndBowTiePattern = "973ph4"
TorsoHpWithDarkRedSweaterAndYellowNeckPattern = "973ph8"
TorsoHpWithGreySweaterAndLightFleshNeckPattern = "973ph7"
TorsoHpWithJacketAndDarkRedSweaterPattern = "973ph6"
TorsoHpWithJacketAndLightGreySweaterPattern = "973ph5"
TorsoHoodedSweatshirtOpenWithDrawstringsOverDarkRedTopPattern = "973p9b"
TorsoHoodieWithWhiteLaces_BlueScrollworkAndKangarooPocketPattern = "973p1v"
TorsoHoodieWithZipperOverLimeAndGreenStripedShirtPattern = "973p99"
TorsoJacketOpenOverRedAndLightGreenTopWithNecklacePattern = "973p1w"
TorsoPirateVest_AnchorTattoo_RopeOnBackPattern = "973p3v"
TorsoPoloShirtWithPocketWithPenAndLegoLogoOnBackPattern = "973p98"
TorsoRiddlerBlackQuestionMarkOnChestAndBackAndLavenderBeltPattern = "973pbn"
TorsoSwBeltPocketsAndNecklaceBlisslFlutePattern = "973prg"
TorsoSwJawaWithDarkBrownPouchesAndBlackAndTanStraps = "973pri"
TorsoSwTankTopPattern = "973prf"
TorsoScarletSpiderPattern = "973pbs"
TorsoShirtWithRedCollarAndSpaceshipOrbitingClassicSpaceHelmetPattern = "973p97"
TorsoSpacesuitWithSilverBeltAndChestPieceWithSpaceShuttleLogoOnFrontAndBackPattern = "973p8s"
TorsoSportsShirtWith_39_OrangeBasketballAndOrangeAndRedDiamondsPattern = "973p9k"
TorsoW_GoldLionAndBeltOnRed_WhiteQuart_Pattern = "973p4l"
TorsoW_LeatherJacket_Badge_Police_BackPattern = "973p7n"
TorsoW_SwLooseDressLightBluishGreyFoldsPattern = "973psy"
TorsoW_SwLooseDressLightGreyFoldsPattern = "973pry"
TorsoW_Shirt_Badge_BlueTie_Police_BackPattern = "973p7m"
TorsoW_White_BlueTriangles_Blue_WhiteAmuletPattern = "973pwc"
TorsoW_White_BlueTriangles_Red_WhiteAmuletPattern = "973pwb"
TorsoWith_S_LogoGrey_BluePattern = "973p23"
TorsoWith_S_LogoRed_BlackPattern = "973p14"
TorsoWith_Tine_Stickers = "973d01"
TorsoWith_1RacingPattern = "973p80"
TorsoWith2ChineseLettersYellowStripePattern = "973paq"
TorsoWith3_PieceSuit_WhiteShirtAndRedTiePattern = "973pdb"
TorsoWithAirGaugeAndPocketPattern = "973p29"
TorsoWithAirplaneLogoPattern = "973p16"
TorsoWithAirplaneLogoAnd_Air_BadgePattern = "973p85"
TorsoWithAirplaneLogo_WhiteShirtAndSuitPattern = "973p8v"
TorsoWithAnchorMotifPattern = "973p09"
TorsoWithAquanautSubLogo_Zipper_AndWeightBeltPattern = "973pag"
TorsoWithAquasharkBlueSharkWithRed_X_Pattern = "973pak"
TorsoWithArcticExplorerLogo_ZipperAndWhiteFurTrimPattern = "973p92"
TorsoWithArcticParka_A2_Pattern = "973p7b"
TorsoWithArcticParkaA1Pattern = "973p7a"
TorsoWithArgyleSweaterAndWhiteShirtCollarPattern = "973p93"
TorsoWithArmorVest_SilverWolfHeadAndBlueRoundJewelPattern = "973paw"
TorsoWithAstroPattern = "973p6f"
TorsoWithAutoroutePattern = "973p27"
TorsoWithAviatorJacketW_Eagle_Smh_OnBackPattern = "973pq5"
TorsoWithBandageAndGoldNecklacePattern = "973pq0"
TorsoWithBandage_GoldNecklaceAndGoldBeltPattern = "973pq1"
TorsoWithBatmanRobinPattern = "973pb2"
TorsoWithBlackCollarAndPocketsPattern = "973pst"
TorsoWithBlackDiagonalZipAndPocketPattern = "973p0b"
TorsoWithBlackDungareesPattern = "973p1a"
TorsoWithBlackFalconPattern = "973p43"
TorsoWithBlackFuturonPattern = "973p6b"
TorsoWithBlackHalfPattern = "973pbm"
TorsoWithBlackStripePattern = "973px2"
TorsoWithBlackSuit_RedShirt_GoldClaspsPattern = "973pdf"
TorsoWithBlacktronIPattern = "973p52"
TorsoWithBlacktronIiPattern = "973p51"
TorsoWithBlouse_BeltAndNecklacePattern = "973pq6"
TorsoWithBlueCastleBodicePattern = "973p4n"
TorsoWithBlueDungareesPattern = "973p1b"
TorsoWithBlueFlowersOnTiedShirtPattern = "973p2g"
TorsoWithBlueFuturonPattern = "973p6c"
TorsoWithBlueHorizontalStripesPattern = "973p1d"
TorsoWithBlueImperialGuardOfficerPattern = "973p3r"
TorsoWithBlueImperialGuardPattern = "973p3n"
TorsoWithBluePlaidShirtPattern = "973pdd"
TorsoWithBlueShirtAndSafetyStripesPattern = "973p8g"
TorsoWithBlueStripedDungareesPattern = "973p1r"
TorsoWithBlueUndershirtGreenBowAndGunPattern = "973pw5"
TorsoWithBlueV_Neck_PocketsAnd3ButtonsPattern = "973p06"
TorsoWithBlueVest_CollarAndStarPattern = "973par"
TorsoWithBlueVest_CollarAndTestTubePattern = "973pas"
TorsoWithBlueVest_Pockets_ShirtAndDrillPattern = "973pat"
TorsoWithBlueVest_Tools_ShirtAndBombPattern = "973pau"
TorsoWithBlueAndMintGreenStripesPattern = "973p2e"
TorsoWithBomberJacket_BlackShirtPattern = "973p70"
TorsoWithBomberJacket_Belt_BlackShirtPattern = "973pa5"
TorsoWithBrownVest_AscotAndBeltPattern = "973p3b"
TorsoWithBrownVest_BuckleAndStringBowtiePattern = "973pw9"
TorsoWithBusinessSuitAndTie_BreathingApparatusAndOctanLogoPattern = "973p6h"
TorsoWithButtonsAndPoliceBadgePlainPattern = "973p1f"
TorsoWithCard_Suit_Vest_AndGoldFobPattern = "973pw8"
TorsoWithCastleBodiceAndCloakPattern = "973p4h"
TorsoWithCastleBreastplatePattern = "973p40"
TorsoWithCastleChainmailPattern = "973p41"
TorsoWithCastleCrossedPikesPattern = "973p42"
TorsoWithCastleFemaleArmourPattern = "973p4g"
TorsoWithCastleRed_GreyPattern = "973p47"
TorsoWithCatsuit_RingZipperPullAndBeltPattern = "973pb8"
TorsoWithChefJacketWith8ButtonsAndLongRedNeckerchiefPattern = "973p8t"
TorsoWithChefPattern = "973p2a"
TorsoWithClassicSpacePattern = "973p90"
TorsoWithClassicSpaceSimulatedWearPattern = "973p91"
TorsoWithClockworkRobotPattern = "973pc67"
TorsoWithConstructionSafetyVestWithReflectiveStripesPattern = "973p8i"
TorsoWithCorsetWithReddishBrownLacesPattern = "973p4w"
TorsoWithCrewNeckSweaterWithBrightPinkCollarPattern = "973pd17"
TorsoWithCrewNeckSweaterWithCollarAnd_HailToTheChef_OnWhiteApronPattern = "973p89"
TorsoWithCrownOnDarkblue_MediumblueQuartersPattern = "973p4j"
TorsoWithDarkBrownMusclesAndBellyPattern = "973px1"
TorsoWithDarkOrangeVestAndPurpleBowTiePattern = "973pb7"
TorsoWithDarkblueMusclesAndGoldNecklacePattern = "973pq2"
TorsoWithDkGreenCorsetWithLeatherBeltPattern = "973p4x"
TorsoWithDkgray_Black_AndYellowBatmanPattern = "973pb1"
TorsoWithDoctorCoatOverMediumAzureUndershirt_Stethoscope_PocketWithPensPattern = "973p9m"
TorsoWithDragonClaspsAndRedTiesFrontAndDragonBackPattern = "973pnb"
TorsoWithDragonHeadPattern = "973p4b"
TorsoWithDressAndRedBeadsNecklacePattern = "973pd13"
TorsoWithDress_WhiteApronAndRedBeadsNecklace = "973p88"
TorsoWithExplorienLogoPattern = "973p55"
TorsoWithExtremeTeamJacketLogoPattern = "973p8a"
TorsoWithFantasyEraPeasantRopeBeltPattern = "973p4v"
TorsoWithFemaleJumpsuitWithZipper_WhiteAndGoldMarkingsAndClassicSpaceLogoPattern = "973p6g"
TorsoWithFemalePiratePattern = "973p38"
TorsoWithFemaleShirtWithTwoButtonsAndShellPendantPattern = "973p8k"
TorsoWithFemaleWhiteShirt_OpenJacketPattern = "973pae"
TorsoWithFireFighterFiveButtonPattern = "973p21"
TorsoWithForestmanBlackCollarPattern = "973p50"
TorsoWithForestmanBlueCollarPattern = "973p49"
TorsoWithForestmanMaroonCollarPattern = "973p48"
TorsoWithForestmanAndPursePattern = "973p46"
TorsoWithFourButtonSuitAndTrainLogoPattern = "973p84"
TorsoWithFrontAndRearStickersWith_Tine_LogoOnBlueBackground = "973d05"
TorsoWithGoldAsianCharactersFrontAndDragonBackPattern = "973pna"
TorsoWithGoldFobAnd100DollarBillsPattern = "973pwa"
TorsoWithGoldIcePlanetPattern = "973p61"
TorsoWithGoldNecklaceAndYellowUndershirtPattern = "973p72"
TorsoWithGoldTiePinAndBlackTiePattern = "973pd7f"
TorsoWithGreenCastleBodicePattern = "973p4q"
TorsoWithGreenDungareesPattern = "973p1j"
TorsoWithGreenFuturonPattern = "973p6i"
TorsoWithGreenLanternLogoOnWhiteBackgroundAndMusclesOutlinePattern = "973pc0"
TorsoWithGreenStripesAndLeatherStrapPattern = "973p3t"
TorsoWithGreyMusclesAndArcReactorPattern = "973pbh"
TorsoWithGreyAndGoldBatmanPattern = "973pb0"
TorsoWithGryffindorUniformPattern = "973ph1"
TorsoWithHpHarryPotterUniformGryffindorStripeAndShieldPattern = "973pha"
TorsoWithHpNarcissaMalfoyPattern = "973ph9"
TorsoWithHalterTopWithGreenApplesAndLimeSpotsPattern = "973p1x"
TorsoWithHogwartsUniformPattern = "973ph0"
TorsoWithHoodedSweatshirt_BluePocket_DrawstringPatt = "973p7j"
TorsoWithHorseHeadPattern = "973p2m"
TorsoWithHospitalEmtStarOfLife_OpenCollar_ButtonsAndPocketPenPattern = "973p8n"
TorsoWithImperialSoldierPattern = "973p3u"
TorsoWithInsectoidsFemale_BlueDiamondUnderCircuitsPattern = "973p58"
TorsoWithInsectoidsRobotPattern = "973p53"
TorsoWithInsectoidsWithBlueXUnderCircuitsPattern = "973p56"
TorsoWithInsectoidsWithGreenCircuitry_SilverHoseOnSides = "973p6k"
TorsoWithInsectoidsWithSilverCircuitry_LightningBoltsOverBlueCircle = "973p67"
TorsoWithInsectoidsWithSilverLogoOnBlackPanels = "973p57"
TorsoWithInsectoidsWithSilverXOnGreenVerniers = "973p59"
TorsoWithIntegralArms = "17"
TorsoWithIronFistPattern = "973pbq"
TorsoWithIronManArmouredSuitClassicPattern = "973pbt"
TorsoWithIronManArmouredSuitMarkViPattern = "973pbb"
TorsoWithIronManArmouredSuitMarkViiPattern = "973pbc"
TorsoWithIronManArmouredSuitMarkXliiPattern = "973pbf"
TorsoWithIronManArmouredSuitMarkXviiPattern = "973pbe"
TorsoWithIronPatriotArmouredSuitPattern = "973pbg"
TorsoWithIronSpiderPattern = "973pbu"
TorsoWithIvyDressPattern = "973pb5"
TorsoWithJacketAndMediumDarkFleshCollarOverLightBluishGreyShirtPattern = "973p9l"
TorsoWithJacketOverLtBlueButtonDownShirtPattern = "973p7z"
TorsoWithJacketOverMediumBlueSkinAndLightBlueWaistPattern = "973pd1d"
TorsoWithJacketOverShirtAndPrisonStripesPattern = "973p7t"
TorsoWithJacketWithPocketsOverDarkRedV_NeckSweaterPattern = "973p1y"
TorsoWithJacket_BadgeAndRedTiePattern = "973p7r"
TorsoWithJacket_BowTie_Vest_BoutonnierePattern = "973pc23"
TorsoWithJacket_OrangeVest_GreenNeck_ChiefPattern = "973pb3"
TorsoWithJacket_PinkShirt_RingOnNecklacePattern = "973p7h"
TorsoWithJacket_Pocket_Badge_BlueTiePattern = "973p7q"
TorsoWithJacket_TieAndPoliceBadgeYellowStarPattern = "973p76"
TorsoWithJacket_WhiteShirtAndDarkPinkTiePattern = "973pd31"
TorsoWithJacket_WhiteShirt_AndNecklacePattern = "973pa8"
TorsoWithJacket_ZipperedPocketsAndClassicSpaceLogoPattern = "973p6j"
TorsoWithJailStripesAnd_23768_Pattern = "973p7k"
TorsoWithLotrArmor_BrownBelt_GoldBucklePattern = "973pm5"
TorsoWithLotrButtonShirtWithBracesPattern = "973pm2"
TorsoWithLotrCloakWithDarkBluishGreyFoldsPattern = "973pm9"
TorsoWithLotrCloak_RopeAndBeltPattern = "973pm1"
TorsoWithLotrCoat_DarkBluishGreyShirt_EvenstarPendantAndBeltPattern = "973pm7"
TorsoWithLotrJacketAndDarkRedVestPattern = "973pm3"
TorsoWithLotrJacketAndYellowVestPattern = "973pm8"
TorsoWithLotrLeatherArmorWithBucklePattern = "973pm0"
TorsoWithLotrOrcLeatherArmorPattern = "973pma"
TorsoWithLotrScaleArmorAndBeltPattern = "973pm6"
TorsoWithLotrShirt_DarkBrownBelt_BagPattern = "973pm4"
TorsoWithLotrUruk_HaiLurtzMusclesOutlinePattern = "973pmb"
TorsoWithLaunchCommandLogoAndEquipmentPattern = "973p1q"
TorsoWithLeatherJacketPattern = "973p28"
TorsoWithLeatherJacketAndBadgePattern = "973p7u"
TorsoWithLeatherJacketAndLightGrayShirtPattern = "973pa9"
TorsoWithLeatherJacket_SlashPocketsAndWaistbandOverRedShirt_SilverOvalBeltBucklePattern = "973p9e"
TorsoWithLifebeltLogoAndIdCardPattern = "973p79"
TorsoWithLightBlueParabolaShapedBellyPattern = "973pcfd"
TorsoWithLightGreyWaistPattern = "973pd1e"
TorsoWithLionGoldOnBlueShieldPattern = "973p4f"
TorsoWithMbaLevel1LogoPattern = "973pt0"
TorsoWithMbaLevel2LogoPattern = "973pt1"
TorsoWithMtronLogoPattern = "973p68"
TorsoWithMagentaQuestionmarkPattern = "973pb6"
TorsoWithMailHornPattern = "973p1s"
TorsoWithMaroon_RedQuartersShieldPattern = "973p4u"
TorsoWithMayanNecklace_TribalShirt_NavelPattern = "973pac"
TorsoWithMedallion_BeltAndSilverButtonsPattern = "973p3d"
TorsoWithMediumBlueTrimAndFlowersPattern = "973p96"
TorsoWithMetallicSilverRomanSegmentedBodyArmourPattern = "973pc6a"
TorsoWithModernFirefighterPatternType1 = "973p77"
TorsoWithModernFirefighterPatternType2 = "973p78"
TorsoWithNecklaceAndBlueSquaresPattern = "973pwe"
TorsoWithNecklaceAndFringePattern = "973pc11"
TorsoWithNeonYellowStripes_RadioAndBadgePattern = "973p7v"
TorsoWithNinjaWrap_GoldShuriken_AndArmourPattern = "973pn6"
TorsoWithNinjaWrap_GoldShuriken_AndDaggerPattern = "973pn7"
TorsoWithNinjaWrap_SilverShuriken_DaggerPattern = "973pn5"
TorsoWithOctan_Oil_BadgePattern = "973p81"
TorsoWithOctanLogoPattern = "973pt2"
TorsoWithOpenJacketOverStripedVestPattern = "973p34"
TorsoWithOpenJacket_BlueTopAndNavelPattern = "973pd1"
TorsoWithOpenSuitJacketWithWhiteShirt_DarkRedTieAndReddishBrownBeltPattern = "973p9i"
TorsoWithOpenSuitRevealingSupermanLogoPattern = "973pbk"
TorsoWithOrangeDungarees_Sweater_BluePenInPocketPattern = "973p8m"
TorsoWithPalmTreePattern = "973p2h"
TorsoWithPalmTreeAndDolphinPattern = "973p2j"
TorsoWithPalmTreeAndHorsePattern = "973p2k"
TorsoWithPatchPocketPattern = "973p26"
TorsoWithPharaohBreastplatePattern = "973pa2"
TorsoWithPinstripeVest_RedTieAndPocketWatchPattern = "973p9h"
TorsoWithPinstripedJacketAndGoldTiePattern = "973p8l"
TorsoWithPirateCaptainPattern = "973p36"
TorsoWithPirateGreenVest_Shirt_AndBeltPattern = "973p3c"
TorsoWithPirateOpenJacketOverBrownVestPattern = "973p39"
TorsoWithPiratePurpleVestAndAnchorTattooPattern = "973p30"
TorsoWithPirateRaggedShirtAndDaggerPattern = "973p3a"
TorsoWithPirateStripesBlue_CreamPattern = "973p32"
TorsoWithPirateStripesBlue_WhitePattern = "973p3e"
TorsoWithPirateStripesRed_BlackPattern = "973p33"
TorsoWithPirateStripesRed_CreamPattern = "973p31"
TorsoWithPirateVestWith3GoldButtonsOverRedAndWhiteStripedShirtPattern = "973p3g"
TorsoWithPlainShirtWithPocketsPattern = "973p7g"
TorsoWithPoliceFemaleShirtWithGoldBadge_PocketAndBlackTiePattern = "973p7l"
TorsoWithPoliceJacket_SilverBadge_4ButtonsAndBeltPattern = "973pcbf"
TorsoWithPrisonerAnd_50380_Pattern = "973p7s"
TorsoWithPurpleGreatcoatPattern = "973phb"
TorsoWithRes_QOrangePocketsAndLogoPattern = "973p8b"
TorsoWithRacingJacketAndTwoStarsPattern = "973p1h"
TorsoWithRacingJacketAndTwoStarsRedPattern = "973p1n"
TorsoWithRearStickerWhite_1_OnTransparentBackground = "973d06"
TorsoWithRed_GrayShirt_PotBelly_JacketPattern = "973paf"
TorsoWithRedCrossPattern = "973p24"
TorsoWithRedCrossAndStethoscopePattern = "973p25"
TorsoWithRedDungareesPattern = "973p1c"
TorsoWithRedFuturonPattern = "973p6d"
TorsoWithRedHorizontalStripesPattern = "973p1e"
TorsoWithRedImperialGuardOfficerPattern = "973p3s"
TorsoWithRedImperialGuardPattern = "973p3q"
TorsoWithRedNecklaceAndBlueUndershirtPattern = "973p71"
TorsoWithRedShirtAndSuitPattern = "973p22"
TorsoWithRedStrapsAndBelt_YellowClaspsAndCordAndMusclesOutlinePattern = "973pby"
TorsoWithRedTruckPattern = "973p1t"
TorsoWithRedUndershirtAndFringePattern = "973pw7"
TorsoWithRedV_NeckAndButtonsPattern = "973p17"
TorsoWithRedVestAndTrainLogoPattern = "973p82"
TorsoWithRedOnBottomAndFringePattern = "973pwd"
TorsoWithRed_PeachQuartersShieldPattern = "973p4t"
TorsoWithRed_WhiteWaist_ZipperAndWhiteStarPattern = "973pbl"
TorsoWithRidingJacketPattern = "973p12"
TorsoWithRobotPattern = "973p63"
TorsoWithRockRaidersJetPattern = "973paj"
TorsoWithRoyalKnightsLionHead_Neck_ChainPattern = "973p4e"
TorsoWithRoyalKnightsLion_HeadShieldPattern = "973p4d"
TorsoWithRumpledShirt_2RevolversInBeltPattern = "973paa"
TorsoWithSwBenKenobiPattern = "973ps6"
TorsoWithSwBlastArmor_GreenPlates_Bl_GreyPattern = "973prb"
TorsoWithSwBlastArmor_GreenPlates_GreyPattern = "973psb"
TorsoWithSwBlastArmor_SilverPlates_Pattern = "973psj"
TorsoWithSwBlastArmor_UncolouredPlates_Pattern = "973prc"
TorsoWithSwCadBanePattern = "973psw"
TorsoWithSwCamouflageSmockPattern = "973psm"
TorsoWithSwClosedShirtAndBrownBeltPattern = "973pr0"
TorsoWithSwDarkRedRobePattern = "973psx"
TorsoWithSwDarthVaderBluishGreyPattern = "973pra"
TorsoWithSwDarthVaderDeathStarPattern = "973psz"
TorsoWithSwDarthVaderGreyPattern = "973ps7"
TorsoWithSwGreyShirt_BeltWithRedBucklePattern = "973pr7"
TorsoWithSwGunganDarkGrey_DarkTanShirtsPattern = "973pr6"
TorsoWithSwHothTrooperPattern = "973psh"
TorsoWithSwImperialGrandMoffPattern = "973pr9"
TorsoWithSwImperialOfficerPattern = "973psq"
TorsoWithSwImperialShuttlePilotPattern = "973psn"
TorsoWithSwJawaPattern = "973pss"
TorsoWithSwJediRobesAndSashPattern = "973ps2"
TorsoWithSwLayeredShirt_BrownBeltPattern = "973pr3"
TorsoWithSwLayeredShirt_BrownBelt_BraidPattern = "973pr4"
TorsoWithSwLayeredShirt_WaistSashPattern = "973ps8"
TorsoWithSwLeiaHothJacketPattern = "973prd"
TorsoWithSwMoistureFarmerPattern = "973psv"
TorsoWithSwNuteGunrayPattern = "973prh"
TorsoWithSwPocket_VestAndTechno_BucklePattern = "973psc"
TorsoWithSwProtocolDroidPattern = "973psr"
TorsoWithSwRebelA_WingPilotPattern = "973ps0"
TorsoWithSwRebelMechanicPattern = "973psa"
TorsoWithSwRebelPilotBluishGreyPattern = "973pr8"
TorsoWithSwRebelPilotPattern = "973ps1"
TorsoWithSwRobeAndRopeBeltPattern = "973pr2"
TorsoWithSwRobe_SashAndSilverNeckClaspPattern = "973pr1"
TorsoWithSwScoutTrooperBluishGreyPattern = "973pre"
TorsoWithSwScoutTrooperGreyPattern = "973pse"
TorsoWithSwShadowTrooperPattern = "973prk"
TorsoWithSwShirtOpenCollarNoVestPattern = "973ps4"
TorsoWithSwStormtrooperPattern = "973psk"
TorsoWithSwTunicAndBeltPattern = "973psf"
TorsoWithSwVest_WhiteShirt_LtFleshNeckPattern = "973pr5"
TorsoWithSwVest_WhiteShirt_YellowNeckPattern = "973ps5"
TorsoWithSwWrap_AroundTunicPattern = "973ps3"
TorsoWithSafariShirt_BlackNeckAndBrownHolsterPattern = "973pa7"
TorsoWithSafariShirt_BlueTee_RedBandanaPattern = "973pa6"
TorsoWithSafariShirt_Gun_RedBandana_WhiteChestPattern = "973pax"
TorsoWithSafariShirt_Gun_RedBandana_YellowChestPattern = "973pa3"
TorsoWithSafariShirt_TanBandana_CompassPattern = "973pq7"
TorsoWithSafetyVestWithReflectiveStripesOverMediumBlueShirt_IdBadgeAndRedPenAroundNeckPattern = "973p8r"
TorsoWithSamuraiDragonRobePattern = "973pn0"
TorsoWithSamuraiRobe_SashAndDaggerPattern = "973pn1"
TorsoWithScaleMailAndRedDiamondAmuletPattern = "973p4c"
TorsoWithSecurityGuardPattern = "973pb4"
TorsoWithShellLogoPattern = "973p60"
TorsoWithSheriffPattern = "973pw4"
TorsoWithShirtCollarAndSandGreenNeckPattern = "973px0"
TorsoWithShirtWith3FlowersOverYellowAndSilverStripePattern = "973p8y"
TorsoWithShirtWithBlackBellyBumpAndCollarOutlinePattern = "973pd11"
TorsoWithShirtWithCargoLogoPattern = "973p2n"
TorsoWithShirt_Pink_SalmonTieAndIdBadge = "973p87"
TorsoWithSilverArmorFrontAndBackAndDrillPattern = "973pav"
TorsoWithSilverIcePlanetPattern = "973p62"
TorsoWithSilverMetalPlatesAndDarkPinkElectricalWiresAndButtonsPattern = "973pbx"
TorsoWithSilverZodEmblemAndOvalBeltBuckleAndMusclesOutlinePattern = "973pbw"
TorsoWithSixButtonSuitAndAirplanePattern = "973p04"
TorsoWithSixButtonSuitAndAnchorPattern = "973p05"
TorsoWithSixButtonSuit_RedTie_AndAirplanePattern = "973p8u"
TorsoWithSkullAndDkpinkRosesPattern = "973pc00"
TorsoWithSkullAndRedRosesPattern = "973pc44"
TorsoWithSlingshotOnBackPattern = "973pd12"
TorsoWithSlytherinUniformPattern = "973ph2"
TorsoWithSoccerShirt_10_And_Zidane_Pattern = "973pg0"
TorsoWithSoccerTeamGoalkeeperAndWhite_1_Pattern = "973pg1"
TorsoWithSoccerTeamRed_BlueStripesAndRed_2Pattern = "973pg2"
TorsoWithSoccerTeamRed_BlueStripesAndRed_3Pattern = "973pg3"
TorsoWithSoccerTeamRed_BlueStripesAndRed_4Pattern = "973pg4"
TorsoWithSoccerTeamRed_BlueStripesAndRed_7Pattern = "973pg5"
TorsoWithSoccerTeamRed_BlueStripesAndRed_9Pattern = "973pg6"
TorsoWithSoccerTeamRed_BlueStripesAndRed_10Pattern = "973pg7"
TorsoWithSoccerTeamRed_BlueStripesAndRed_11Pattern = "973pg8"
TorsoWithSoccerTeamRed_BlueStripesAndRed_18Pattern = "973pg9"
TorsoWithSoccerTeamWhiteStripesAndWhite_2Pattern = "973pga"
TorsoWithSoccerTeamWhiteStripesAndWhite_3Pattern = "973pgb"
TorsoWithSoccerTeamWhiteStripesAndWhite_4Pattern = "973pgc"
TorsoWithSoccerTeamWhiteStripesAndWhite_7Pattern = "973pgd"
TorsoWithSoccerTeamWhiteStripesAndWhite_9Pattern = "973pge"
TorsoWithSoccerTeamWhiteStripesAndWhite_10Pattern = "973pgf"
TorsoWithSoccerTeamWhiteStripesAndWhite_11Pattern = "973pgg"
TorsoWithSoccerTeamWhiteStripesAndWhite_18Pattern = "973pgh"
TorsoWithSpacePoliceIiChiefPattern = "973p6a"
TorsoWithSpacePoliceIiAndRadioPattern = "973p69"
TorsoWithSpacePortLogo_TubeAnd_C1_Pattern = "973p8e"
TorsoWithSpacesuitWithFrontStrap_BadgesAndPortsPattern = "973p5b"
TorsoWithSpiderManPattern = "973pbv"
TorsoWithSpider_GirlPattern = "973pbp"
TorsoWithSpottedSingletAndNecklacePattern = "973p2f"
TorsoWithSpyriusPattern = "973p66"
TorsoWithStethoscopeAndPocketWithPensPattern = "973p86"
TorsoWithStickerShellLogoOnWhiteBackground = "973d04"
TorsoWithStraightZipperJacketPattern = "973p13"
TorsoWithStraplessSuntopPattern = "973p2c"
TorsoWithStripedShirtAndSilverButtonsPattern = "973p3f"
TorsoWithStuddedArmourPattern = "973p45"
TorsoWithStudiosDirectorPattern = "973pd0"
TorsoWithSuitAndRedTieWithTrainLogoPattern = "973p83"
TorsoWithSuitAndTiePattern = "973p18"
TorsoWithSuitWithPocket_RedTie_TrainPattern = "973p8h"
TorsoWithSupermanLogoAndDarkBlueMusclesPattern = "973pb9"
TorsoWithSuspendersAndRedBowtiePattern = "973pa1"
TorsoWithSweaterV_NeckOverLightBlueShirtPattern = "973p94"
TorsoWithSweaterWithDarkAzureShirtWithCollarAndOrangeAscotTiePattern = "973px3"
TorsoWithT_ShirtWith4WhiteHorizontalStripesFrontAndBackWithCatPattern = "973p9n"
TorsoWithT_Shirt_BlueDungareesAndToolsPattern = "973p8z"
TorsoWithTvLogoPatternLarge = "973p1m"
TorsoWithTvLogoPatternSmall = "973p1k"
TorsoWithTankTop_Braces_YellowSkinPattern = "973pq4"
TorsoWithTankTop_Stains_Wrench_AndTattooPattern = "973pab"
TorsoWithTelephonePattern = "973p8f"
TorsoWithThreeWhiteAndSilverSquaresPattern = "973p8x"
TorsoWithTownVestWithPocketsAndStripedTiePattern = "973p8j"
TorsoWithTrainLogoAndRedDiagonalStripesPattern = "973p19"
TorsoWithTri_ColouredShieldLargePattern = "973p4r"
TorsoWithTri_ColouredShieldAndGoldTrimPattern = "973p4s"
TorsoWithUfoAlienCircuitryAndRedLeverPattern = "973p6w"
TorsoWithUfoAlienInsigniaAnd3BlueBarsPattern = "973p6v"
TorsoWithUfoAlienOrangeAndSilverPattern = "973p54"
TorsoWithUfoAlienSilverAndGoldCircuitryPattern = "973p6x"
TorsoWithUfoAlienTriangularInsigniaPattern = "973p6u"
TorsoWithUsCavalryGeneralPattern = "973pw1"
TorsoWithUsCavalryOfficerPattern = "973pw2"
TorsoWithUsCavalrySoldierPattern = "973pw3"
TorsoWithUnitronPattern = "973p64"
TorsoWithV_NeckShirtAndBlueOverallsPattern = "973p7w"
TorsoWithVerticalRedStripesAnd_Clutchers_Pattern = "973pc3g"
TorsoWithVerticalStripedBlue_RedPattern = "973p02"
TorsoWithVerticalStripedRed_BluePattern = "973p01"
TorsoWithVestWithPatchPocketsPattern = "973p73"
TorsoWithVest_PatchPocketsAndPoliceBadgeYellowStarPattern = "973p74"
TorsoWithVest_SlingshotInBeltPattern = "973pad"
TorsoWithVoldemortRobePattern = "973ph3"
TorsoWithWaistcoat_WhiteShirtAndBandolierPattern = "973pq3"
TorsoWithWaiterPattern = "973p20"
TorsoWithWarMachineArmouredSuitPattern = "973pbd"
TorsoWithWhiteBracesAndCartridgeBeltPattern = "973pw6"
TorsoWithWhiteButtonsAndPoliceBadgePlainSticker = "973d03"
TorsoWithWhiteDiagonalZipAndPocketPattern = "973p0a"
TorsoWithWhiteEnvelopeAndZipperPattern = "973p7x"
TorsoWithWhiteOrcaBellyPattern = "973pd6e"
TorsoWithWhiteRope_PatchedCollarPattern = "973pdg"
TorsoWithWhiteScarfAndBrownBeltPattern = "973pay"
TorsoWithWhiteShirtAndJacketPattern = "973p03"
TorsoWithWhiteShirtWithDarkBlueTie_GoldAirlinePin_BeltAndIdBadgePattern = "973p8w"
TorsoWithWhiteSuit_BrownVestAndTiePattern = "973pa4"
TorsoWithWhiteTigerPattern = "973pbr"
TorsoWithWhiteVerticalStripesAndRedBowTiePattern = "973p7i"
TorsoWithWhite_RedArmourAndWhiteBelt = "973pwf"
TorsoWithWhite_Yellow_OrangeRacesuitPattern = "973p95"
TorsoWithWindsurfboardPattern = "973p2d"
TorsoWithWolfpackPattern = "973p44"
TorsoWithYellowButtonsAndGreyBeltSticker = "973d02"
TorsoWithYellowChestAndWhiteBeadsNecklacePattern = "973pd14"
TorsoWithYellowFuturonPattern = "973p6e"
TorsoWithZipperJacketAnd3PocketsPattern = "973p1u"
TorsoWithZipperJacketAndPoliceLogoPattern = "973p75"
TorsoWithZipperAndPoliceBadgePlainPattern = "973p1g"



#!/usr/bin/env python
"""
Takes care of downloading the complete.zip from LDraw.org
And generating the parts.lst from it
"""
from __future__ import print_function
import zipfile
import os
from distutils.dir_util import copy_tree

import appdirs
from mklist.generate import generate_parts_lst

from ldraw.compat import urlretrieve
from ldraw.dirs import get_data_dir
from ldraw.utils import ensure_exists

LDRAW_URL = 'http://www.ldraw.org/library/updates/complete.zip'


def download_main(parts_lst_path):
    """ download complete.zip, mklist, main function"""
    tmp_ldraw = ensure_exists(appdirs.user_cache_dir('pyldraw'))

    if not os.path.exists(parts_lst_path):
        retrieved = os.path.join(tmp_ldraw, "complete.zip")

        print('retrieve the complete.zip from ldraw.org ...')
        urlretrieve(LDRAW_URL, filename=retrieved)

        print('unzipping the complete.zip ...')
        zip_ref = zipfile.ZipFile(retrieved, 'r')
        zip_ref.extractall(tmp_ldraw)
        zip_ref.close()

        output_dir = ensure_exists(os.path.abspath(os.path.join(parts_lst_path, '..')))

        copy_tree(os.path.join(tmp_ldraw, 'ldraw'), os.path.join(output_dir))

        print('mklist...')
        generate_parts_lst('description',
                           os.path.join(output_dir, 'parts'),
                           parts_lst_path)


if __name__ == '__main__':
    download_main(os.path.join(get_data_dir(), 'ldraw', 'parts.lst'))



"""
Some utils functions
"""

import collections
import re
import os


def clean(input_string):
    """
    Cleans a description string:

    :param input_string:
    :return:
    """
    input_string = re.sub(r'\W+', '_', input_string)
    input_string = re.sub(r'_x_', 'x', input_string)
    return input_string


def camel(input_string):
    """ Returns a CamelCase string """
    return ''.join(x for x in input_string.title() if not x.isspace())


def ensure_exists(path):
    """ makes the directory if it does not exist"""
    try:
        os.makedirs(path)
    except OSError:
        pass
    return path


# https://stackoverflow.com/a/6027615
def flatten(input_dict, parent_key='', sep='.'):
    """ flatten a dictionary """
    items = []
    for key, value in input_dict.items():
        new_key = parent_key + sep + key if parent_key else key
        if isinstance(value, collections.MutableMapping):
            items.extend(flatten(value, new_key, sep=sep).items())
        else:
            items.append((new_key, value))
    return dict(items)



#!/usr/bin/env python
"""
Called by ldraw.library_gen to generate the ldraw/library/colours.py file
"""
from __future__ import print_function
import codecs
import os

import pystache

from ldraw.resources import get_resource
from ldraw.utils import clean, camel


def get_c_dict(colour):
    """
    Gets a dict from a Colour object
    """
    return {'code': colour.code,
            'full_name': colour.name,
            'name': camel(clean(colour.name)),
            'alpha': colour.alpha,
            'rgb': colour.rgb, 'colour_attributes': colour.colour_attributes}


def gen_colours(parts, output_dir):
    """
    Generates a colours.py from library data
    """
    print('generate ldraw.library.colours...')

    colours_mustache = get_resource(os.path.join('templates', 'colours.mustache'))
    colours_template_file = codecs.open(colours_mustache, 'r', encoding='utf-8')
    colours_template = pystache.parse(colours_template_file.read())

    context = {'colours': [get_c_dict(c) for c in parts.colours_by_name.values()]}
    context['colours'].sort(key=lambda r: r['code'])

    colours_str = pystache.render(colours_template, context=context)
    library_path = os.path.join(output_dir, 'library')

    colours_py = os.path.join(library_path, 'colours.py')

    with codecs.open(colours_py, 'w', encoding='utf-8') as generated_file:
        generated_file.write(colours_str)



""" ldraw.library generation """
import os
import shutil
import hashlib

from ldraw.dirs import get_data_dir
from ldraw.generation.colours import gen_colours
from ldraw.generation.parts import gen_parts
from ldraw.parts import Parts
from ldraw.utils import ensure_exists


LIBRARY_INIT = """\"\"\" the ldraw.library module, auto-generated \"\"\"
__all__ = [\'colours\']
"""


def library_gen_main(parts_lst, output_dir):
    """ main function for the library generation """
    library_path = os.path.join(output_dir, 'library')
    ensure_exists(library_path)
    hash_path = os.path.join(library_path, '__hash__')

    md5_parts_lst = hashlib.md5(open(parts_lst, 'rb').read()).hexdigest()

    if os.path.exists(hash_path):
        md5 = open(hash_path, 'r').read()
        if md5 == md5_parts_lst:
            return

    parts = Parts(parts_lst)

    library__init__ = os.path.join(library_path, '__init__.py')

    with open(library__init__, 'w') as library__init__:
        library__init__.write(LIBRARY_INIT)
    shutil.copy('ldraw-license.txt', os.path.join(library_path, 'license.txt'))

    gen_colours(parts, output_dir)
    gen_parts(parts, output_dir)

    open(hash_path, 'w').write(md5_parts_lst)


if __name__ == '__main__':
    library_gen_main(os.path.join(get_data_dir(), 'ldraw', 'parts.lst'), 'ldraw')



# pylint: disable=too-few-public-methods, too-many-arguments

""" classes for lines in parts path """


class OptionalLine(object):
    """ an optional Line """

    def __init__(self, colour, point1, point2, point3, point4):
        self.colour = colour
        self.point1 = point1
        self.point2 = point2
        self.point3 = point3
        self.point4 = point4


class Quadrilateral(object):
    """ a quadrilateral """

    def __init__(self, colour, point1, point2, point3, point4):
        self.colour = colour
        self.point1 = point1
        self.point2 = point2
        self.point3 = point3
        self.point4 = point4

    @property
    def points(self):
        """ returns the points array """
        return [self.point1, self.point2, self.point3, self.point4]


class Line(object):
    """ a 3D line """

    def __init__(self, colour, point1, point2):
        self.colour = colour
        self.point1 = point1
        self.point2 = point2

    @property
    def points(self):
        """ returns the points array """
        return [self.point1, self.point2]


class Triangle(object):
    """ a triangle """

    def __init__(self, colour, point1, point2, point3):
        self.colour = colour
        self.point1 = point1
        self.point2 = point2
        self.point3 = point3

    @property
    def points(self):
        """ returns the points array """
        return [self.point1, self.point2, self.point3]


class MetaCommand(object):
    """ a metacommand """

    def __init__(self, text):
        self.text = text


class Comment(object):
    """ a comment """

    def __init__(self, text):
        self.text = text



"""
svg.py - An SVG writer for the ldraw Python package.

Copyright (C) 2010 David Boddie <david@boddie.org.uk>

This file is part of the ldraw Python package.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from ldraw.geometry import Vector2D
from ldraw.writers.common import Writer


class Polygon(object):
    """Polygon used for SVG rendering"""

    # pylint: disable=too-few-public-methods

    def __init__(self, zmin, points, colour, piece):
        self.zmin = zmin
        self.points = points
        self.colour = colour
        self.piece = piece

    def __lt__(self, other):
        return self.zmin < other.zmin


SVG_PREAMBLE = """<?xml version="1.0" standalone="no"?>
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"
  "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">
<svg width="{svg_args.width:.6f}cm" \
height="{svg_args.height:.6f}cm" \
viewBox="{view1:.6f} {view2:.6f} {svg_args.width:.6f} {svg_args.height:.6f}" \
xmlns="http://www.w3.org/2000/svg" version="1.1">
"""

SVG_LINE = """<line stroke="{rgb}" \
stroke-width="{stroke_width}" \
opacity="{opacity:.6f}" \
x1="{point1.x:.6f}" \
y1="{point1.y:.6f}" \
x2="{point2.x:.6f}" \
y2="{point2.y:.6f}" />
"""

SVG_POLYGON_PREAMBLE = """<polygon fill="{rgb}" \
stroke="{stroke_colour}" \
stroke-width="{stroke_width}" \
opacity="{opacity:.6f}" \
points=\""""

POLYGON_FORMAT = '<polygon fill="%s" points="%.6f,%.6f %.6f,%.6f %.6f,%.6f %.6f,%.6f" />\n'


class SVGArgs(object):
    """Data-only container for arguments passed to an SVG writer"""

    # pylint: disable=too-few-public-methods

    def __init__(self, width, height, stroke_colour=None,
                 stroke_width=None, background_colour=None):
        # pylint: disable=too-many-arguments
        self.width = width
        self.height = height
        self.stroke_colour = stroke_colour
        self.stroke_width = stroke_width
        self.background_colour = background_colour


def _project_polygons(width, height, polygons):
    # vx' = width + az
    # vy' = height + bz
    pixel_x = 0.5
    pixel_y = 0.5
    half_width = width / 2.0
    half_height = height / 2.0
    new_polygons = []
    for polygon in polygons:
        new_points = []
        for point in polygon.points:
            point_x = half_width * (point.x / (half_width + pixel_x * -point.z))
            point_y = half_height * (point.y / (half_height + pixel_y * -point.z))
            new_point = Vector2D(point_x, point_y)
            new_points.append(new_point)
        new_polygons.append((new_points, polygon))
    return new_polygons


def write_preamble(args, svg_file):
    """ write the preamble and polygon def """
    svg_file.write(SVG_PREAMBLE.format(view1=0.0, view2=0.0, svg_args=args))
    if args.background_colour is not None:
        arguments = (args.background_colour,
                     0.0, 0.0, args.width, 0.0,
                     args.width, args.height, 0.0, args.height)
        svg_file.write(POLYGON_FORMAT % arguments)


class SVGWriter(Writer):
    """Writes a model into a SVG"""

    # pylint: disable=too-few-public-methods
    def write(self, model, svg_file, svg_args):
        """Writes the SVG """
        polygons = self._polygons_from_objects(model)
        polygons.sort()
        shapes = _project_polygons(svg_args.width, svg_args.height, polygons)
        self._write(shapes, svg_file, svg_args)

    def _write(self, shapes, svg_file, args):
        stroke_width = args.stroke_width if args.stroke_width else "0.1%"

        write_preamble(args, svg_file)

        shift = Vector2D(args.width / 2.0, args.height / 2.0)
        for points, polygon in shapes:
            rgb = self.parts.colours.get(polygon.colour, "#ffffff")
            stroke_colour = args.stroke_colour if args.stroke_colour else rgb
            context = dict(rgb=rgb,
                           stroke_width=stroke_width,
                           stroke_colour=stroke_colour,
                           opacity=self._opacity_from_colour(polygon.colour))
            if len(points) == 2:
                context['point1'] = Vector2D(points[0].x, -points[0].y) + shift
                context['point2'] = Vector2D(points[1].x, -points[1].y) + shift

                svg_file.write(SVG_LINE.format(**context))
            else:
                svg_file.write(SVG_POLYGON_PREAMBLE.format(**context))
                for point in points:
                    dpoint = Vector2D(point.x, -point.y) + shift
                    svg_file.write('{point.x:.6f},{point.y:.6f} '.format(point=dpoint))
                svg_file.write('" />\n')
        svg_file.write("</svg>\n")
        svg_file.close()

    def _line_get_poly(self, obj,
                       top_level_piece,
                       current):
        camera_position = self.camera_position

        points = [current.matrix * p + current.position - camera_position for p in obj.points]

        return self._common_get_poly(obj, top_level_piece, current.colour, points)

    def _get_polygon(self, top_level_piece, colour, projections):  # pylint: disable=no-self-use
        return [Polygon(min(p.z for p in projections),
                        projections,
                        colour, top_level_piece)]



""" data and config directories """

import os
import appdirs

from ldraw.utils import ensure_exists


def get_data_dir():
    """ get the directory where to put some data """
    return ensure_exists(
        os.environ.get('LDRAW_DATA_DIR', appdirs.user_data_dir('pyldraw')))


def get_config_dir():
    """ get the directory where the config is """
    return ensure_exists(appdirs.user_config_dir('pyldraw'))



#!/usr/bin/env python

"""
ldr2svg.py - An LDraw to SVG convertor tool.

Copyright (C) 2010 David Boddie <david@boddie.org.uk>

This file is part of the ldraw Python package.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import argparse

from ldraw.tools import (widthxheight, vector_position,
                         get_model, get_coordinate_system,
                         verify_camera_look_at)
from ldraw.writers.svg import SVGArgs


def main():
    """ ldr2svg main function """
    description = """Converts the LDraw file to a SVG file.
    
The viewport size is specified as a pair of floating point numbers representing
lengths in LDraw scene coordinates separated by an \"x\" character.

The camera and look-at positions are x,y,z argument in LDraw scene coordinates
where each coordinate should be specified as a floating point number.

The optional sky background colour is an SVG colour, either specified as
#rrggbb or as a named colour.

"""
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('ldraw_file')
    parser.add_argument('svg_file')
    parser.add_argument('viewport_size', type=widthxheight)
    parser.add_argument('camera_position', type=vector_position)
    parser.add_argument('look_at_position', type=vector_position,
                        required=False, default=vector_position("0,0,0"))
    parser.add_argument('--sky')

    args = parser.parse_args()
    svg_args = SVGArgs(args.viewport_size[0],
                       args.viewport_size[1],
                       background_colour=args.background_colour)

    ldr2svg(args.ldraw_file, args.svg_file,
            args.camera_position, args.look_at_position, svg_args)


def ldr2svg(ldraw_path, svg_path, camera_position, look_at_position, svg_args): # pylint: disable=too-many-arguments
    """ ldr2svg actual implementation """
    verify_camera_look_at(camera_position, look_at_position)

    model, parts = get_model(ldraw_path)

    system = get_coordinate_system(camera_position, look_at_position)

    with open(svg_path, "w") as svg_file:
        from ldraw.writers.svg import SVGWriter
        writer = SVGWriter(camera_position, system, parts)
        writer.write(model, svg_file, svg_args)


if __name__ == "__main__":
    main()



"""
png.py - A PNG writer for the ldraw Python package.

Copyright (C) 2010 David Boddie <david@boddie.org.uk>

This file is part of the ldraw Python package.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
# pylint: disable=no-name-in-module, too-few-public-methods, too-many-locals, too-many-branches
from PIL import Image, ImageColor
import numpy
from PIL.ImageDraw import Draw

from ldraw.geometry import Vector, Vector2D
from ldraw.writers.common import Writer
from ldraw.writers.geometry import Z_MAX, Edge


class PNGArgs(object):
    """ Args to pass to a PNG writer"""

    def __init__(self, distance, image_size, stroke_colour=None, background_colour=None):
        """
        :param distance: distance of the camera
        :param image_size: size of the image as a string (e.g. '800x800')
        :param stroke_colour: colour of the edges
        :param background_colour: colour of the background
        """
        self.distance = distance
        self.image_size = image_size
        self.stroke_colour = stroke_colour + (255, )
        self.background_colour = background_colour


class PNGWriter(Writer):
    """
    Renders a LDR model into a PNG
    """

    # pylint: disable=too-few-public-methods
    def write(self, model, png_path, png_args):
        """
        Writes the model's polygons to the provided PNG file

        :param model: LDR model
        :type model: Part
        :param png_path: where to output the PNG
        :param png_args: Arguments for the rendering (distance, etc.)
        :type png_args: PNGArgs

        :return:
        """
        distance = png_args.distance
        image_size = png_args.image_size
        stroke_colour = png_args.stroke_colour
        background_colour = png_args.background_colour
        args = ('RGBA', tuple(image_size))
        if background_colour is not None:
            args += (background_colour,)
        image = Image.new(*args)

        depth = numpy.empty((image_size[0], image_size[1]), "f")
        depth[:] = 1 << 32 - 1
        polygons = self._polygons_from_objects(model)

        viewport_scale = min(float(image_size[0]), float(image_size[1]))
        # Draw opaque polygons first.
        for polygon in polygons:
            if polygon.alpha == 1.0:
                polygon.project(distance)
                polygon.render(image, depth, viewport_scale, stroke_colour)
        # Draw translucent polygons last.
        for polygon in polygons:
            if polygon.alpha < 1.0:
                polygon.project(distance)
                polygon.render(image, depth, viewport_scale, stroke_colour)
        image.save(png_path)

    def _get_polygon(self, top_level_piece, colour, projections):
        rgb = self.parts.colours.get(colour, "#ffffff")
        alpha = self._opacity_from_colour(colour)
        return [Polygon(projections, rgb, alpha)]


class Polygon(object):
    """
    Describes a polygon for PNG rendering
    """

    def __init__(self, points, rgb, alpha):
        self.points = points
        self.rgb = ImageColor.getrgb(rgb)
        self.alpha = alpha
        self.rgba = self.rgb + (int(255 * alpha),)
        self.projected = []

    def project(self, distance):
        """
        project to screen
        px/c = x/(c + z)
        px = c * x / (c + z)
        """
        for point in self.points:
            self.projected.append(
                Vector((distance * point.x) / (distance + -point.z),
                       (distance * point.y) / (distance + -point.z), -point.z)
            )

    def render(self, image, depth, viewport_scale, stroke_colour):
        """
        Sort the edges of the polygon by their minimum projected y
        coordinates, discarding horizontal edges.
        """
        edges = self.get_edges(image, viewport_scale)
        if not edges:
            return
        width, height = image.size

        end_py = edges[-1].point2.y
        if end_py < 0:
            return

        edge1 = edges.pop(0)

        if edge1.point1.y >= height:
            return

        edge2 = edges.pop(0)

        int_edge1_y1 = int(edge1.point1.y)

        if int_edge1_y1 < edge1.point1.y or int_edge1_y1 < edge2.point1.y:
            int_edge1_y1 += 1
        while int_edge1_y1 <= end_py and int_edge1_y1 < height:
            # Retrieve new edges as required.
            if int_edge1_y1 >= edge1.point2.y:
                if not edges:
                    break
                edge1 = edges.pop(0)
            if int_edge1_y1 >= edge2.point2.y:
                if not edges:
                    break
                edge2 = edges.pop(0)
            if int_edge1_y1 < 0:
                int_edge1_y1 += 1
                continue
            # Calculate the starting and finishing coordinates of the span
            # at the current y coordinate.
            gradient_1 = Vector2D(edge1.dx_dy, edge1.dz_dy)
            start_point_1 = Vector2D(edge1.point1.x, edge1.point1.z)
            start_1 = start_point_1 + (int_edge1_y1 - edge1.point1.y) * gradient_1

            gradient_2 = Vector2D(edge2.dx_dy, edge2.dz_dy)
            start_point_2 = Vector2D(edge2.point1.x, edge2.point1.z)
            start_2 = start_point_2 + (int_edge1_y1 - edge2.point1.y) * gradient_2

            # Do not render the span if it lies outside the image or has
            # values that cannot be stored in the depth buffer.
            # Truncate the span if it lies partially within the image.
            if start_1.x > start_2.x:
                start_2, start_1 = start_1, start_2
            # Only calculate a depth gradient for the span if it is more than
            # one pixel wide.
            if start_1.y <= 0 and start_2.y <= 0:
                int_edge1_y1 += 1
                continue
            elif start_1.y >= Z_MAX and start_2.y >= Z_MAX:
                int_edge1_y1 += 1
                continue

            start_x, end_sx = int(start_1.x), int(start_2.x)
            if start_x < start_1.x:
                start_x += 1
            if start_x >= width:
                int_edge1_y1 += 1
                continue
            elif end_sx < 0:
                int_edge1_y1 += 1
                continue

            int_edge1_y1 = self.draw_span(depth,
                                          end_sx,
                                          image,
                                          int_edge1_y1,
                                          start_1,
                                          start_2,
                                          start_x,
                                          stroke_colour, width)

    def get_edges(self, image, viewport_scale):
        """
        get edges for rendering
        """
        width, height = image.size
        edges = []
        len_points = len(self.points)
        for i in range(len_points):
            point1 = self.projected[i].copy()
            point1.x = width / 2 + (point1.x * viewport_scale)
            point1.y = height / 2 - (point1.y * viewport_scale)
            j = (i + 1) % len_points

            point2 = self.projected[j].copy()
            point2.x = width / 2 + (point2.x * viewport_scale)
            point2.y = height / 2 - (point2.y * viewport_scale)
            # Append the starting and finishing y coordinates, the starting
            # x coordinate, the dx/dy gradient of the edge, the starting
            # z coordinate and the dz/dy gradient of the edge.
            if int(point1.y) < int(point2.y):
                edges.append(Edge(point1, point2))
            elif int(point1.y) > int(point2.y):
                edges.append(Edge(point2, point1))
        edges.sort(key=lambda e: e.sort_key)
        return edges

    # pylint: disable=too-many-arguments
    def draw_span(self,
                  depth, end_sx,
                  image, int_edge1_y1,
                  start_1, start_2,
                  start_x, stroke_colour,
                  width):
        """ draw a span """
        if start_1.x != start_2.x:
            start_dz_dx = (start_2.y - start_1.y) / (start_2.x - start_1.x)
        else:
            start_dz_dx = 0.0
        if start_x < 0:
            start_x = 0
        if end_sx >= width:
            end_sx = width - 1
        # Draw the span.
        while start_x <= end_sx:
            start_z = start_1.y + start_dz_dx * (start_x - start_1.x)
            if 0 < start_z <= depth[int(start_x)][int(int_edge1_y1)]:
                if self.alpha < 1.0:
                    pixel = image.getpixel((start_x, int_edge1_y1))
                    rgb = int((1 - self.alpha) * pixel[0] + self.alpha * self.rgb[0]), \
                          int((1 - self.alpha) * pixel[1] + self.alpha * self.rgb[1]), \
                          int((1 - self.alpha) * pixel[2] + self.alpha * self.rgb[2]), 255
                    image.putpixel((start_x, int_edge1_y1), rgb)
                else:
                    depth[int(start_x)][int(int_edge1_y1)] = start_z
                    image.putpixel((start_x, int_edge1_y1), self.rgba)
            start_x += 1
        if stroke_colour:
            if 0 <= start_1.x < width and 0 < start_1.y <= depth[int(start_1.x)][int(int_edge1_y1)]:
                image.putpixel((int(start_1.x), int_edge1_y1), stroke_colour)
            if 0 <= start_2.x < width and 0 < start_2.y <= depth[int(start_2.x)][int(int_edge1_y1)]:
                image.putpixel((int(start_2.x), int_edge1_y1), stroke_colour)
        int_edge1_y1 += 1
        return int_edge1_y1



""" Some geometry elements used in Writers """


class Edge(object):
    # pylint:disable=too-few-public-methods
    """
    Holds an edge of a polygon, used during pixel rendering
    """

    def __init__(self, point1, point2):
        """
        :param point1:
        :type point1: Vector
        :param point2:
        :type point2: Vector
        """

        self.point1 = point1
        self.point2 = point2

        self.dx_dy = (point2.x - point1.x) / (point2.y - point1.y)
        self.dz_dy = (point2.z - point1.z) / (point2.y - point1.y)

    @property
    def sort_key(self):
        """ used for sorting the edges before rendering """
        return (self.point1.y, self.point2.y,
                self.point1.x, self.dx_dy,
                self.point1.z, self.dz_dy)


Z_MAX = 1 << 16



#!/usr/bin/env python
"""
Generates the ldraw.library.parts namespace
"""
from __future__ import print_function
import codecs
import os
import itertools

from progress.bar import Bar
import pystache

from ldraw.resources import get_resource
from ldraw.utils import clean, camel, ensure_exists, flatten
from ldraw.parts import PartError

PARTS__INIT__TEMPLATE = get_resource(os.path.join('templates', 'parts__init__.mustache'))
PARTS__INIT__TEMPLATE = codecs.open(PARTS__INIT__TEMPLATE, 'r', encoding='utf-8')
PARTS__INIT__TEMPLATE = pystache.parse(PARTS__INIT__TEMPLATE.read())

PARTS_TEMPLATE = get_resource(os.path.join('templates', 'parts.mustache'))
PARTS_TEMPLATE = codecs.open(PARTS_TEMPLATE, 'r', encoding='utf-8')
PARTS_TEMPLATE = pystache.parse(PARTS_TEMPLATE.read())

SECTION_SEP = '|'


def write_section_file(parts_dir, list_of_parts, mod_path):
    """ Writes a single section files"""
    list_of_parts.sort(key=lambda o: o['description'])
    part_str = pystache.render(PARTS_TEMPLATE, context={'parts': list_of_parts})
    parts_py = os.path.join(parts_dir, mod_path)
    ensure_exists(os.path.dirname(parts_py))
    with codecs.open(parts_py, 'w', encoding='utf-8') as generated_file:
        generated_file.write(part_str)


def get_part_dict(parts, description):
    """ Gets a dict context for a part """

    try:
        code = parts.parts_by_name[description]
        #part = parts.part(code=code)
        return {
         #   'path': part.path,
            'description': description,
            'class_name': clean(camel(description)),
            'code': code
        }
    except (PartError, KeyError):
        return {}


def gen_parts(parts, output_dir):
    """
    Generates the ldraw.library.parts namespace
    :param parts: Parts object
    :param output_dir: where to output the library
    :return:
    """
    print('generate ldraw.library.parts, this might take a long time...')

    library_path = os.path.join(output_dir, 'library')
    parts_dir = ensure_exists(os.path.join(library_path, 'parts'))

    sections = _get_sections(parts)

    packages = _get_packages(sections)

    for package_name, modules in packages.items():
        generate_parts__init__(library_path, modules, package_name)

    for section_name, section_parts in sections.items():
        generate_section(parts, parts_dir, section_name, section_parts)


def _get_packages(sections):
    packages = {}
    for section_name in sections:
        rsplitted = section_name.rsplit(SECTION_SEP, 1)
        package = {'module_name': rsplitted[-1]}
        if len(rsplitted) > 1:
            packages.setdefault(rsplitted[0], []).append(package)
        else:
            packages.setdefault('', []).append(package)
    return packages


def _get_sections(parts):
    sections = {}
    for key, value in flatten(parts.parts, sep=SECTION_SEP).items():
        rsplitted = key.rsplit(SECTION_SEP, 1)
        description = rsplitted[-1]
        sections.setdefault(rsplitted[0], {})[description] = value
    return sections


def module_path(section):
    """ gets the .py path from a section name"""
    return section.replace(SECTION_SEP, os.sep) + '.py'


def generate_section(parts, parts_dir, section_name, section_parts):
    """ generate all the sections in ldraw.library.parts namespace"""
    parts_list = []
    progress_bar = Bar('section %s ...' % section_name, max=len(section_parts))
    for description in section_parts:
        parts_list.append(get_part_dict(parts, description))
        progress_bar.next()
    progress_bar.finish()
    parts_list = [x for x in parts_list if x != {}]
    if section_name == 'others':
        parts_list.sort(key=lambda r: r.get('category', 'others'))
        for name, grouped in itertools.groupby(parts_list, key=lambda r: r.get('category', 'others')):
            grouped = list(grouped)
            if name is None:
                name = 'others'

            write_section_file(parts_dir, grouped, module_path(name))
    else:
        write_section_file(parts_dir, parts_list, module_path(section_name))


def generate_parts__init__(library_path, modules, package_name):
    """ generate the appropriate __init__.py to make submodules in ldraw.library.parts """
    parts__init__str = pystache.render(PARTS__INIT__TEMPLATE, context={'sections': modules})
    if package_name == '':
        parts__init__ = os.path.join(library_path, 'parts', '__init__.py')
    else:
        parts__init__ = os.path.join(library_path, 'parts', package_name, '__init__.py')
    ensure_exists(os.path.dirname(parts__init__))
    with codecs.open(parts__init__, 'w', encoding='utf-8') as parts__init__file:
        parts__init__file.write(parts__init__str)



#!/usr/bin/env python

"""
ldr2inv.py - Create parts lists from LDraw models.

Copyright (C) 2012 David Boddie <david@boddie.org.uk>

This file is part of the ldraw Python package.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import codecs
import sys
import argparse

from ldraw.lines import Comment
from ldraw.pieces import Piece
from ldraw.tools import get_model


def main():
    """ Main function for ldr2inv """
    parser = argparse.ArgumentParser()
    parser.add_argument('ldraw_file')
    parser.add_argument('output')

    args = parser.parse_args()

    ldr2inv(args.ldraw_file, args.output)


def ldr2inv(ldraw_path, inventory_path):
    """ Actual ldr2inv implementation """
    model, parts = get_model(ldraw_path)

    inventory = {}
    length = 0

    for obj in model.objects:
        if not isinstance(obj, Piece):
            continue

        name = ""
        subpart = parts.part(code=obj.part)
        if subpart:
            for component in subpart.objects:
                if isinstance(component, Comment):
                    name = component.text
                    break
            else:
                sys.stderr.write("No name information for part: %s\n" % obj.text)

            inventory[name] = inventory.get(name, 0) + 1
            length = max(len(name), length)

    length += (4 - (length % 4))

    try:
        with codecs.open(inventory_path, 'w', encoding='utf-8') as inv_file:
            items = list(inventory.items())
            items.sort()

            for name, number in items:
                padding = " " * (length - len(name))
                inv_file.write("%s%s%i\n" % (name, padding, number))

    except IOError:
        sys.stderr.write("Failed to write inventory file: %s\n" % inventory_path)
        sys.exit(1)





if __name__ == "__main__":
    main()



# coding=utf-8
"""
library/parts.py - Auto-generated Part classes for the Python ldraw package.

Copyright (C) 2018 Matthieu Berthomé <rienafairefr@gmail.com>

This file is part of the ldraw Python package.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.


<< This file is auto-generated, you shouldn't have to modify it>>
"""
# pylint: disable=invalid-name, line-too-long, too-many-lines

AcousticGuitar = "25975"
AcousticGuitarWithBlackFingerboardAndNeckAndMetallicSilverStringsPattern = "25975p01"
AcousticGuitarWithBlackTuningPegsAndMetallicSilverStringsPattern = "25975p02"
Airtanks = "3838"
ArmorSamurai = "30174"
ArmorShoulderPadsWith1StudOnFront_2StudsOnBack = "11097"
ArmorShoulderPadsWithNeckProtection_1StudOnFront_2StudsOnBack = "15086"
ArmorShoulderPadsWithRidges = "11438"
ArmourPlate = "2587"
ArmourPlateWithLotrKingTheodenPattern = "2587pm0"
ArmourSwCloneTrooperPauldron = "61190b"
ArmourSwCloneTrooperTasset = "61190c"
ArmourShoulder_NeedsWork_ = "30121"
ArrowQuiver = "4498"
Axe = "95330"
AxeWithTwin_Blade = "95052"
BabyBody = "25126"
BabyBodyWithBodysuit = "15526"
BabyBodyWithYellowHandsPattern = "25126p01"
BabyBodyWithYellowHandsAndBibWithElephantPattern = "25126p02"
BackpackNon_Opening = "2524"
BackpackOpenableClosed = "30158c01"
BackpackOpenableOpen = "30158"
BackpackSkydiver = "12897"
BackpackWithHorizontalPinHoles = "15406"
BackpackWithSleepingBag = "30323"
Bandana = "30133"
BarWithBatWings = "98721"
BarbedWireLoop = "62700"
Basketball = "43702"
BasketballNetForHoop = "43374"
BasketballWithBlackSpokesPattern = "43702p01"
BatWing = "98722"
Battleaxe = "3848"
BattleaxeBlade = "53454"
BeardLongForked = "60750"
BeardLongWithFiveBraids = "60749"
BeardMediumShort = "93223"
BeardShort = "15501"
BeardWithRoundedEnd = "10052"
Bicycle_Complete_ = "4719c01"
BicycleFrame = "4719"
BicycleWithIntegralTyreWheels = "4719c03"
BinocularsWithRoundEyepiece = "30162"
BladedClawSpread = "10187"
Book = "33009-f1"
BookBaseWithTwoStuds = "24324"
BookCover = "24093"
BookCoverWithDisneyCastleAndGoldTrimPattern = "24093p01"
BookOpen = "33009-f2"
BookWithBookCoverMoriaChroniclesStickers = "33009dm0"
Bottle1X1X2Cylindrical = "95228"
Bottle1X1X2CylindricalWithBottleShipPattern = "95228p01"
BowWithArrow = "4499"
BreastplateWithFinsAndEmbossedAtlantisTriangle = "89917"
Broom = "4332"
Bugle = "71342"
CameraMovie = "30148"
CameraSnapshotType1 = "30089a"
CameraSnapshotType2 = "30089b"
CameraWithSideSight = "4360"
CandleFlame = "37775"
Cannon2X4Base = "2527"
Cannon2X8Non_Shooting = "518"
CannonNon_ShootingWithBase_Shortcut_ = "2527c01"
CannonShooting_Complete_ = "2533c01"
CannonShootingBarrelAndBreech = "2535c01"
CapAviator = "30171"
CapAviatorGoggles = "30170"
CapAviatorWithBlackGoggles = "30171c01"
CapWithLongFlatPeak = "4485"
CapWithShortArchedPeak = "86035"
CapWithShortArchedPeakAndBlackHeadphones_EarProtection = "11303c01"
CapWithShortArchedPeakWithSeamsAndTopPinHole = "11303"
Cape = "4524"
CapeCloth = "50231"
CapeCloth_FormedFloating_ = "50231c02"
CapeCloth_Formed_ = "50231c01"
CapeClothScalloped5Points = "56630"
CapeClothScalloped5Points_Formed_ = "56630c01"
CapeClothShort = "99464"
CapeClothShort_Formed_ = "99464c01"
CapeClothWithHolesAndTatteredEdges = "86038"
CapeClothWithHolesAndTatteredEdges_Formed_ = "86038c01"
CapeClothWithInsideDarkRedPattern_Formed_ = "50231p01c01"
CapeClothWithPointedEndsAndCollar = "42450"
CapeClothWithPointedEndsAndCollar_Formed_ = "42450c01"
CastleHelmetWithChin_Guard = "3896"
CastleHelmetWithNeckProtector = "3844"
Cauldron = "4341"
Chain6LWith6LSpanCatenary = "92338-f2"
Chain6LWith6LSpanStraight = "92338-f1"
Chain16LWith16LSpanStraight = "60169"
Chain17LWith17LSpanStraight = "30104"
ChainEndForChain6LAnd17L = "30104k01"
ChainEndForChain16L = "60169k01"
ChainLinkForChain6LAnd17L = "30104k02"
ChainLinkForChain16L = "60169k02"
ChainsawBlade = "6117"
CheerleaderPomPom = "87997"
CheerleaderPomPomWithBluePattern = "87997p01"
CheerleaderPomPomWithRedPattern = "87997p02"
CircularBladeSaw = "30194"
Cleaver = "98369"
ClothWrap_BandanaWithRoundedTop = "18927"
CoinWith_1_GothicType = "96904"
CoinWith_2_GothicType = "96905"
CoinWith_5_GothicType = "96906"
CoinWith_10_GothicType = "96907"
CoinWith_10_Sans_SerifType = "70501a"
CoinWith_20_Sans_SerifType = "70501b"
CoinWith_30_Sans_SerifType = "70501c"
CoinWith_40_Sans_SerifType = "70501d"
Comb = "30112b"
Compass_Complete_ = "889c01"
ComputerLaptop_Closed_ = "62698-f3"
ComputerLaptop_OpenFlat_ = "62698-f1"
ComputerLaptop_Open_ = "62698-f2"
ConicalFlask = "u9180"
ConicalFlaskTransclearWithColouredBase = "u9180c01"
ConstructionHelmet = "3833"
ConstructionHelmetWithHairShort = "16175"
ConstructionHelmetWithMediumDarkFleshHairShortPattern = "16175p02"
ConstructionHelmetWithReddishBrownHairShortPattern = "16175p01"
ContainerD_Basket = "4523"
Crossbow = "2570"
CrownTiara = "33322"
CrownTiaraFlatWithArchedBottom = "96491"
CrownWithBar0_5L = "25516"
Cup = "3899"
CupTakeaway = "15496"
CupWithBarHandle = "38014"
Dagger = "88288"
DaggerWithPearlLightGrayBlade = "88288p01"
DinnerPlate = "6256"
DiverMask = "30090"
DynamiteSticksBundle = "64728"
EagleWing = "93250"
EagleWingWithTanFeathersAndRed_GoldPattern = "93250pq0"
ElectricGuitar = "93564"
ElectricGuitarClassic = "11640"
ElectricGuitarWithBlackStrings_DkpinkLightningPattern = "93564p02"
ElectricGuitarWithSilverStrings_WhiteBodyPattern = "93564p01"
Epaulette = "2526"
FeatheredHeaddressSmall = "87696"
FeathersWithClip = "30127"
FeathersWithClipAndBlackTipPattern = "30127p01"
FeathersWithPin = "30126"
FeathersWithPinAndBlackTipOnOneSidePattern = "30126p02"
FeathersWithPinAndBlackTipOnBothSidesPattern = "30126p01"
FeathersWithPinAndRedTipWithBlackStripePattern = "30126p03"
FireHelmet = "3834"
FireHelmetBreathingHose = "6158"
FireHelmetWithFireLogoShieldPattern = "3834p01"
Flame7LWithBar0_8 = "85959"
Flame7LWithBar0_8WithMarbledTransBrightGreenFlameTipPattern = "85959p05"
Flame7LWithBar0_8WithMarbledTransClearFlameTipPattern = "85959p04"
Flame7LWithBar0_8WithMarbledTransGreenFlameTipPattern = "85959p03"
Flame7LWithBar0_8WithMarbledTransOrange_TransYellowFlameTipPattern = "85959p01"
Flame7LWithBar0_8WithMarbledTransPurpleFlameTipPattern = "85959p06"
Flame7LWithBar0_8WithMarbledTransYellowFlameTipPattern = "85959p02"
FlameWithPegs = "6126a"
FlameWithRim = "6126b"
Flipper = "2599"
FlippersOnSprue = "2599c01"
FoodApple = "33051"
FoodBanana = "33085"
FoodBreadBun = "25386"
FoodCarrot = "33172"
FoodCarrotTop = "33183"
FoodCarrotWithBright_GreenLeaves_Complete_ = "33172c02"
FoodCarrotWithGreenLeaves_Complete_ = "33172c01"
FoodCherry = "22667"
FoodCroissant = "33125"
FoodFrenchBread = "4342"
FoodIceCreamCone = "33120"
FoodIceCreamScoops = "6254"
FoodPieWithCreamFilling = "93568"
FoodPieWithWhiteCreamFillingPattern = "93568p01"
FoodPieWithYellowCreamFillingPattern = "93568p02"
FoodPopsicle = "30222"
FoodPretzel = "10170"
FoodSausage = "33078"
FoodTurkeyBody = "33048"
FoodTurkeyWithLegs_Complete_ = "33048c01"
ForestmanCap = "4506"
ForestmanCapWithSmallRedPlume_Complete_ = "4506c01"
Frypan = "4528"
GhostShroud = "2588"
Goblet = "2343"
GobletWithHollowStem = "6269"
Gollum = "10057"
GollumWithNarrowEyesAndHalf_OpenMouthPattern = "10057pm1"
GollumWithWideEyesAndOpenMouthPattern = "10057pm0"
GrapplingHookWithBar = "15534"
GunBlasterWithStudsOnSides_BottomAndFront = "44709"
GunFlintlockPistol = "2562"
GunLaserKryptonian = "13952"
GunLaserPistol = "87993"
GunLongBlaster = "57899"
GunMusket = "2561"
GunPistolTwoBarrel = "95199"
GunRevolver = "30132"
GunRifle = "30141"
GunSwSmallBlasterDc_17 = "61190a"
GunSemiautomaticPistol = "55707a"
GunShootingBlaster = "15391"
GunShootingBlasterTrigger = "15392"
GunShootingBlasterWithDarkBluishGreyTrigger = "15391c01"
GunShootingBlasterWithDarkBluishGreyTriggerAndTransOrangeProjectile = "15391c02"
GunShortBlaster = "58247"
GunSmallBlasterWithTelescopicSight = "92738"
HairBackslick = "64798"
HairBeard = "6132"
HairBubbleStyle_Afro_ = "87995"
HairCoiled = "21778"
HairCombedFrontToRear = "92081"
HairFemaleLongStraightWithLeftSideParting = "92083"
HairFemalePonytailWithLongFrenchBraid = "88286"
HairFemaleShortSweptSideways = "20877"
HairFemaleWithElaborateKnotBun = "13251"
HairFemaleWithPigtails = "3625"
HairFlatTop = "30608"
HairIslander = "6025"
HairLongStraight = "40239"
HairLongTousledWithCentreParting = "20595"
HairLongWavy = "40251"
HairLongAndHalfBraided_NeedsWork_ = "30475"
HairLongWithCurls = "11255"
HairLongWithFrenchBraid = "59363"
HairLongWithHeadband = "30114"
HairLongWithHeadbandAndFeathers_Complete_ = "30114c01"
HairLongWithOneFrontLock = "85974"
HairMale = "3901"
HairMid_LengthTousledWithCentreParting = "88283"
HairMohawk = "93563"
HairObiWan = "61196"
HairOrcWithDarkTanEarsPattern = "10066p01"
HairOrcWithEars = "10066"
HairOrcWithMediumDarkFleshEarsPattern = "10066p02"
HairOrcWithOliveGreenEarsPattern = "10066p03"
HairPeaked = "42444"
HairPonytail = "6093a"
HairPonytailOff_Centre = "15427"
HairPonytailOff_CentreWithMagentaStripePattern = "15427p02"
HairPonytailOff_CentreWithMagentaAndMediumAzureStripesPattern = "15427p01"
HairPonytailWithLongBangs = "62696"
HairPulledBack = "92756"
HairRockerWithSideburns = "98371"
HairShortBobCut = "62711"
HairShortBowlCut = "40240"
HairShortCombedSidewaysWithLeftParting = "99930"
HairShortSweptBackWithSideburnsAndWidow_SPeak = "21268"
HairShortTousled = "40233"
HairShortWavyWithCentreParting = "26139"
HairShort_TousledWithSideParting = "62810"
HairShort_WavyWithSidePart = "11256"
HairShoulderLength = "4530"
HairSpiked = "98385"
HairSpikyLong = "53982"
HairSpikyShort = "53981"
HairStraightCutWithShortPonytail = "17630"
HairSweptBackIntoBun = "99240"
HairSweptBackTousled = "61183"
HairSweptBackWithWidow_SPeak_Chin_LengthAndBushyInBack = "21269"
HairSweptLeftTousled = "23186"
HairSweptRightWithFrontCurl = "98726"
HairTousled = "10048"
HairTousledAndLayered = "92746"
HairTousledWithSideParting = "87991"
HairWavy = "43751"
HairWavyWindblownToOneSide = "32602"
HairWith2LocksOnLeftSide = "15443"
HairWithSnakes = "12889"
HairWithTwoBuns = "30409"
HairbrushWithLongHandle = "3852a"
HairbrushWithShortHandle = "3852b"
Handcuffs = "61482"
Handlebars = "30031"
HandlebarsWithAngularHandles = "98397"
Harpoon = "57467"
HarpoonWithSmoothBar = "18041"
HeaddressAnubisGuard = "93248"
HeaddressAnubisGuardW_GoldEyeAndStripesPattern = "93248pq0"
HeaddressAztecBird = "99243"
HeaddressAztecBirdWithEyesAndCheeksPattern = "99243p01"
HeaddressBatman = "55704"
HeaddressCatwoman = "98729"
HeaddressCatwomanShortWithWiderApartEyeholes = "55705"
HeaddressHorus = "93249"
HeaddressHorusWithEyePattern = "93249pq0"
HeaddressIndianBuffaloHorned = "30113"
HeaddressIndianWarbonnet = "30138"
HeaddressIndianWarbonnetWithColouredFeathersPattern = "30138p01"
HeaddressNemesType1 = "30168b"
HeaddressNemesType2 = "90462b"
HeaddressNemesWith2DarkblueSnakesPattern = "90462bpq0"
HeaddressNemesWithDarkblueStripesPattern = "90462bpq1"
HeaddressSwZabrakHorns = "92761"
HeaddressTurban = "40235"
HeaddressWerewolf = "42443"
HeaddressWithHookedBeakAndFeathers = "30276"
Headphones_EarProtection = "14045"
HelmetAlienSkullWithFangs = "85945"
HelmetArmy = "87998"
HelmetArmyWithWhiteStencilCrossPattern = "87998p01"
HelmetBubbleHalf = "61287"
HelmetCap = "60748"
HelmetCapWithWings = "60747"
HelmetCastleRohanWithCheekProtection_Comb = "10054"
HelmetCastleRohanWithCheekProtection_CombWithEomerPattern = "10054p01"
HelmetCastleRohanWithCheekProtection_CombWithTheodenPattern = "10054p02"
HelmetCastleUruk_HaiWithLateralComb = "10051"
HelmetCastleW_DragonCrownTopBlackW_DragonPlume = "6122c01"
HelmetCastleW_DragonCrownTopW_CattlehornWhite = "6122c02"
HelmetCastleWithAngledCheekProtection = "48493"
HelmetCastleWithDragonCrownTop = "6122"
HelmetCastleWithFixedFaceGrille = "4503"
HelmetClassicWithCrackedThinChinGuard = "16599"
HelmetClassicWithThickChinGuardAndVisorDimples = "3842b"
HelmetClassicWithThinChinGuardAndVisorDimples = "3842a"
HelmetDarthVader = "30368"
HelmetGladiator = "95676"
HelmetImperialAt_StPilot = "57900"
HelmetIronMan = "10907"
HelmetIronManWithVisorWithGoldFace_WhiteEyesPattern = "10907c01"
HelmetMining = "30325"
HelmetModern = "2446"
HelmetModernWithBlack_BlueAndSilverUnitronPattern = "2446p51"
HelmetModernWithBlueAndSilverSpyriusPattern = "2446p50"
HelmetModernWithOneRedAndTwoGreenStripesPattern = "2446p01"
HelmetMorion = "30048"
HelmetPersian = "88284"
HelmetRangefinderForSwCloneTrooperHelmet = "63586"
HelmetRomanSoldier = "98366"
HelmetSwRebelPilotWithCheckeredPattern = "30370ps4"
HelmetSwRebelPilotWithYellowGridOnBluishgreyPattern = "30370ps1"
HelmetSwRebelPilotWithYellowGridOnGreyPattern = "30370ps5"
HelmetSamurai = "30175"
HelmetSamuraiHorn = "529"
HelmetSamuraiWithHorn_Complete_ = "30175c01"
HelmetScoutTrooper = "30369"
HelmetSkateboard = "46303"
HelmetStarWarsRebelPilot = "30370"
HelmetStarWarsRebelPilotWithBlueRebelLogoPattern = "30370ps3"
HelmetStarWarsRebelPilotWithRedRebelLogoPattern = "30370ps2"
HelmetStormtrooper = "30408"
HelmetStormtrooperWithStormtrooperPattern = "30408p01"
HelmetStormtrooperWithTie_FighterPilotPattern = "30408p02"
HelmetUnderwater = "87754"
HelmetVikingHorn = "53451"
HelmetVikingWithNoseProtector = "53450"
HelmetVikingWithWhiteHorns_Complete_ = "53450c01"
HelmetVisor = "2447"
HelmetVisorIcePlanet = "6119"
HelmetVisorIronMan = "10908"
HelmetVisorIronManWithGoldFace_BlueEyesPattern = "10908p02"
HelmetVisorIronManWithGoldFace_WhiteEyesPattern = "10908p01"
HelmetVisorIronManWithSilverFaceAndDkredLinePattern = "10908p04"
HelmetVisorIronManWithSilverFace_WhiteEyesPattern = "10908p03"
HelmetVisorLargeWithTrapezoidAreaOnTop = "89159"
HelmetVisorSpace = "769"
HelmetVisorUnderwater = "6090"
HelmetVisorWithGrilleAndFeather = "3843"
HelmetWelding = "13792"
HelmetWeldingWithDarkBlueStripesOnBlackVisorPattern = "13792p01"
HelmetWithBatWings = "30105"
HelmetWithChinstrapAndWideBrim = "30273"
HelmetWithFrontProngs = "10305"
HelmetWithFrontProngsWithDarkPurpleHighlightPattern = "10305p01"
HelmetWithHexagonalTopAndGoldAlienPattern = "30120p01"
HelmetWithHexagonalTopAndHoses = "30120"
HelmetWithHexagonalTopAndSilverAlienPattern = "30120p02"
Hips = "3815b"
HipsAndLegs = "73200-f1"
HipsAndLegsShortWithHole = "41879b"
HipsAndLegsShortWithHorizontalStripe = "16709"
HipsAndLegsShortWithYellowHorizontalStripePattern = "16709p01"
HipsAndLegsShortWithYellowHorizontalStripeAndSlingshotOnBackPattern = "16709p02"
HipsAndLegsShortWithoutHole = "41879a"
HipsAndLegsWithBeltAndReflectiveStripes_Emmet_BadgePattern = "3815c8h"
HipsAndLegsWithBeltAndReflectiveStripes_Emmet_BadgeAndSilverTrianglesOnFeetPattern = "3815c8i"
HipsAndLegsWithBlue_WhiteTriangles_FringePattern = "3815cw2"
HipsAndLegsWithButtonedPocketsPattern = "3815ca3"
HipsAndLegsWithClockworkRobotPattern = "3815cc67"
HipsAndLegsWithDarkGreyBelt_DarkRedLoincloth_WhiteClaws_FurTailPattern = "3815caw"
HipsAndLegsWithDarkTurquoise_WhiteTriangles_WhiteFringePattern = "3815cw3"
HipsAndLegsWithDarkblueLoinclothPattern = "3815cq0"
HipsAndLegsWithDecoratedLoinclothPattern = "3815cq1"
HipsAndLegsWithDkblue_GoldLoinclothPattern = "3815cq2"
HipsAndLegsWithGoldCircuitryPattern = "3815c6w"
HipsAndLegsWithGrassSkirtPattern = "3815c3j"
HipsAndLegsWithGreenKilt_ToesPattern = "3815ca2"
HipsAndLegsWithHarlequinRed_BlackPattern = "3815cba"
HipsAndLegsWithHolsterAndBeltPattern = "3815ca9"
HipsAndLegsWithIronManArmouredSuitMarkViPattern = "3815cbb"
HipsAndLegsWithIronManArmouredSuitMarkViiPattern = "3815cbc"
HipsAndLegsWithIronManArmouredSuitMarkXliiPattern = "3815cbf"
HipsAndLegsWithIronManArmouredSuitMarkXviiPattern = "3815cbe"
HipsAndLegsWithIronPatriotArmouredSuitPattern = "3815cbg"
HipsAndLegsWithLotrCoat_ShirttailsAndBeltPattern = "3815cm2"
HipsAndLegsWithLotrLeatherArmorPattern = "3815cm0"
HipsAndLegsWithLotrScaleArmourPattern = "3815cm1"
HipsAndLegsWithLaboratorySmockPattern = "3815cde"
HipsAndLegsWithLeatherArmourPattern = "3815c4f"
HipsAndLegsWithOrangeCablePattern = "3815c6u"
HipsAndLegsWithPurpleGreatcoatPattern = "3815chb"
HipsAndLegsWithRedRomanTunicAndBeltPattern = "3815cc6a"
HipsAndLegsWithRed_WhiteTriangles_FringePattern = "3815cw1"
HipsAndLegsWithRobotPattern = "3815c63"
HipsAndLegsWithSwGunbeltPattern = "3815cs5"
HipsAndLegsWithSwStormtrooperPattern = "3815csk"
HipsAndLegsWithSalmonCablePattern = "3815c6v"
HipsAndLegsWithStuddedBeltAndSafetyPinsPattern = "3815cc44"
HipsAndLegsWithWarMachineArmouredSuitPattern = "3815cbd"
HipsAndLegsWithWhiteApronPattern = "3815c89"
HipsAndLegsWithWhiteAndGoldMarkingsPattern = "3815c6g"
HipsAndLegsWithWoodenLeg = "3815c36"
HipsAndLegs_Sitting = "73200-f2"
HipsWithBlackBeltWithSilverBeltBucklePattern = "3815p8i"
HipsWithClockworkRobotPattern = "3815pc67"
HipsWithDarkGreyBeltWithSilverBuckleAndRivetsPattern = "3815paw"
HipsWithGoldBeltAndDarkblueLoinclothPattern = "3815pq0"
HipsWithGoldBeltAndOrangeCablePattern = "3815p6u"
HipsWithGoldO_BeltAndLoinclothPattern = "3815pq2"
HipsWithGoldX_BeltAndDecoratedLoinclothPattern = "3815pq1"
HipsWithHarlequinBlackSquarePattern = "3815pba"
HipsWithHolsterAndBeltPattern = "3815pa9"
HipsWithIronManArmouredSuitMarkXliiPattern = "3815pbf"
HipsWithIronManArmouredSuitMarkXviiPattern = "3815pbe"
HipsWithIronPatriotArmouredSuitPattern = "3815pbg"
HipsWithLotrCoat_ShirttailsAndBeltPattern = "3815pm2"
HipsWithLotrDarkRedBelt_ScaleArmorPattern = "3815pm1"
HipsWithLotrLeatherArmorPattern = "3815pm0"
HipsWithLaboratorySmockPattern = "3815pde"
HipsWithLeatherBelt_RedStuds_Pattern = "3815p4f"
HipsWithPurpleGreatcoatPattern = "3815phb"
HipsWithRedRomanTunicAndBeltPattern = "3815pc6a"
HipsWithRedTiedSashPattern = "3815bpc1"
HipsWithRed_WhiteTrianglesPattern = "3815pw1"
HipsWithRobotPattern = "3815p63"
HipsWithSwGunbeltPattern = "3815ps5"
HipsWithSwStormtrooperPattern = "3815psk"
HipsWithSilverBeltPattern = "3815p6w"
HipsWithSilverBeltAndSalmonCablePattern = "3815p6v"
HipsWithStuddedBeltPattern = "3815pc44"
HipsWithTentacles = "87749"
HipsWithWarMachineArmouredSuitPattern = "3815pbd"
HipsWithWhiteApronPattern = "3815p89"
HipsWithWhiteAndGoldMarkingsPattern = "3815p6g"
HockeyStickWith1_1LBar = "93559"
Hood = "30381"
HoodFur_Lined = "30287"
HoodFur_LinedWithWhiteFurTrimPattern = "30287p01"
HoodMedievalCowl = "4505a"
HoseNozzleWithSideStringHole = "58367"
HoseNozzleWithSideStringHoleSimplified = "60849"
Hoverboard = "21271"
IceAxe = "30193"
IceAxeWith3_RibHandle = "18738"
Jackhammer = "30228"
Jester_SCap = "62537"
Jester_SCapWithBlackHalfAndWhitePomsPattern = "62537p02"
Jester_SCapWithBlueHalfAndBluePomPattern = "62537p01"
Jester_SCapWithRedHalfAndRedPomPattern = "62537p03"
JetPackWith2OctagonalNozzles_NeedsWork_ = "6023"
Jet_PackWithStudOnFront = "4736"
Key = "62808"
Knife = "44658"
KnifeWithFlatHiltEnd = "95054"
KnifesOnSprue = "44658c01"
Ladle = "4337"
Lance = "3849"
Lantern = "37776"
Lasso = "99253"
LegsOld = "15"
LifeRing = "30340"
LifeguardFloat = "90395"
Lifevest = "2610"
Lightning = "59233"
LightsaberChromeSilver_1SideOn = "577bc01"
LightsaberChromeSilver_2SidesOn = "577bc02"
LightsaberHiltShortCurved = "61199"
LightsaberHiltWithBend = "42114"
LightsaberHiltWithBottomRing = "577b"
LightsaberHiltWithoutBottomRing = "577a"
LightsaberWithBendChromeSilver_1SideOn = "42114c01"
LightsaberWithBendChromeSilver_2SidesOn = "42114c02"
LongBowWithArrow = "93231"
Loudhailer = "4349"
LoudhailerWithOrangeStripePattern = "4349p01"
Machete = "29109"
MachineGunWithDrumMagazine = "55707b"
MagicWand = "6124"
Maracas = "90508"
MaracasWithGreenBorderPattern = "90508p01"
Martian_Complete_ = "30529c01"
MartianLegs = "30530"
MaskWolf = "11233"
MaskWolfWithFangs_ScarsAndWhiteEarsPattern = "11233p01"
MechanicalLegs = "30376"
MechanicalSwBattleDroid_Shortcut_ = "30375cs0"
MechanicalSwBattleDroidCommander_Shortcut_ = "30375cs1"
MechanicalSwBattleDroidPilot_Shortcut_ = "30375cs3"
MechanicalSwBattleDroidSecurity_Shortcut_ = "30375cs2"
MechanicalShoulderWithTwoBar_Holes = "22392"
Megaphone = "39144"
MetalDetector = "4479a"
Microphone = "90370"
MicrophoneWithMetallicSilverTop = "90370p01"
Mop = "24085"
Moustache = "15439"
Mug = "33054"
NinjaWrap = "30177"
OilLamp = "98383"
PaintBrush = "93552"
PaintBrushWithSilverRingAndGreenTipPattern = "93552p01"
PaintPalette = "93551"
PaintPaletteWithPaintDropsPattern = "93551p01"
PaintRollerBrushHandle = "12885"
Pharaoh_SStaffWithForkedEnd = "93252"
Pickaxe = "3841"
PipeWrench_NeedsWork_ = "4328"
Pitchfork = "4496"
PlumeDragon = "87687"
PlumeDragonWingLeft = "87685"
PlumeDragonWingRight = "87686"
PlumeHorsetailWithBar0_5L = "22411"
PlumeLarge = "87694"
PlumeSmall = "87693"
PlumeTriple = "87692"
Plume_FlameTriple = "64647"
Plunger = "11459"
PlungerWithMediumDarkFleshHandle = "11459p01"
PolearmHalberd = "6123"
PoliceHelmet = "13789"
PoliceHelmetWithSilverBadgePattern = "13789p01"
PotSmallWithHandleHolders = "98374"
PropellorWithBatWings = "55707e"
Pushbroom = "3836"
RadioWithLongHandle = "3962b"
RadioWithShortHandle = "3962a"
RiceBowl = "34172"
Ring1X1 = "11010"
RingWithTriangle = "87748"
RingWithTriangleWithGoldBandsAndCrabPattern = "87748p01"
RingWithTriangleWithGoldBandsAndMantaRayPattern = "87748p05"
RingWithTriangleWithGoldBandsAndSharkPattern = "87748p03"
RingWithTriangleWithGoldBandsAndSquidPattern = "87748p04"
RingWithTriangleWithGoldBandsAndTrianglePattern = "87748p06"
RingWithTriangleWithGoldBandsAndTurtlePattern = "87748p02"
RollerSkate = "11253"
RussianFurCap = "43894"
Sack = "92590"
Satchel = "61976"
Saucepan = "4529"
SaucerWithHollowStudOnTop = "38799"
Saxophone = "13808"
SaxophoneWithBlackMouthpiece = "13808p01"
ScabbardForTwoSwords = "88290"
ScabbardForTwoSwordsWithBar = "19141"
Scooter = "15396c01"
ScooterBody = "15396"
ScooterWithRider = "15396c02"
ScubaTank = "30091"
Seat2X2 = "4079"
ServingTray2X3 = "29635"
Sextant = "30154"
Shackles = "55707d"
ShieldBroadWithSpikedBottomAndCutoutCorner = "10049"
ShieldOctagonalWithStud = "48494"
ShieldOctagonalWithTrollSkullOnDarkRedPattern = "61856p40"
ShieldOctagonalWithoutStud = "61856"
ShieldOval = "92747"
ShieldOvalWithSwGunganPatrolShieldPattern = "92747p01"
ShieldOvoid = "2586"
ShieldOvoidW_CrownOnDark_MedBlueQuartersPattern = "2586p4j"
ShieldOvoidW_DkgreenDragonOnMddkflesh_TanPattern = "2586p4k"
ShieldOvoidWithAmericanIndianPattern = "2586pw1"
ShieldOvoidWithBatPattern = "2586p4f"
ShieldOvoidWithBlueDragonPattern = "2586p4c"
ShieldOvoidWithGallopingHorseAndFeathersPattern = "2586pw2"
ShieldOvoidWithGoldLionOnRed_WhiteQuart_Pattern = "2586p4l"
ShieldOvoidWithGoldenLionPattern = "2586ph1"
ShieldOvoidWithGreenDragonPattern = "2586p4b"
ShieldOvoidWithIndigoIslandersPattern = "2586p30"
ShieldOvoidWithRoyalKnightsLionPattern = "2586p4d"
ShieldOvoidWithSwGunganPatrolShieldPattern = "2586ps1"
ShieldOvoidWithSilverSkullOnDarkRedPattern = "2586p4h"
ShieldOvoidWithSilverSnakePattern = "2586ph2"
ShieldPentagonalCastellated = "22408"
ShieldPentagonalJagged = "22409"
ShieldRectangularCurvedWithStud = "98367"
ShieldRound = "3876"
ShieldRoundBowed = "75902"
ShieldRoundBowedWithBullseyeWithStarPattern = "75902p02"
ShieldRoundBowedWithDkgreenAndGoldRohanPattern = "75902p01"
ShieldRoundBowedWithGoldEaglePattern = "75902p03"
ShieldRoundBowedWithMercedes_BenzLogo = "75902p04"
ShieldRoundBowedWithSwDejarikHologameBoardPattern = "75902ps1"
ShieldRoundFlat = "59231"
ShieldRoundFlatWithHelmetOnDark_MediumBlueQuartersPattern = "59231p40"
ShieldRoundFlatWithSilverSkullOnDarkRedPattern = "59231p4h"
ShieldRoundType2 = "91884"
ShieldRoundType2WithAztecBirdOnDarkRedPattern = "91884p04"
ShieldRoundType2WithDarkBrownRingAnd4RivetsPattern = "91884p01"
ShieldRoundType2WithDragonHeadsAndOrnamentsPattern = "91884p03"
ShieldRoundType2WithSilverRivetsPattern = "91884p02"
ShieldScarab = "93251"
ShieldTriangular = "3846"
ShieldTriangularLong = "18836"
ShieldTriangularLongWithBlackPyramidAndSunPattern = "18836pq0"
ShieldTriangularW_CrownOnDk_MedBlueQuart_Pattern = "3846p4j"
ShieldTriangularW_GoldLionOnRed_WhiteQuart_Patt_ = "3846p4l"
ShieldTriangularW_WhiteMalteseCrossOnRedSticker = "3846d02"
ShieldTriangularW_YellowTrefoilsOnDkblueSticker = "3846d07"
ShieldTriangularWithBatPattern = "3846p4f"
ShieldTriangularWithBlackFalconBlueBorderPattern = "3846p45"
ShieldTriangularWithBlackFalconYellowBorderPattern = "3846p46"
ShieldTriangularWithBlueDragonPattern = "3846p4c"
ShieldTriangularWithBlueLionOnYellowPattern = "3846p4g"
ShieldTriangularWithCrownOnDark_PinkSticker = "3846d01"
ShieldTriangularWithCrownOnVioletSticker = "3846d05"
ShieldTriangularWithDragonOnMddkflesh_TanPattern = "3846p4k"
ShieldTriangularWithForestmanPattern = "3846p48"
ShieldTriangularWithGreenChevronsOnYellowSticker = "3846d03"
ShieldTriangularWithGreenChevronsOnYellow_Ltgray = "3846d06"
ShieldTriangularWithLionHead_Blue_YellowPattern = "3846p4e"
ShieldTriangularWithMaroon_RedQuartersPattern = "3846p4u"
ShieldTriangularWithRedCrossAndHelmetPattern = "3846p01"
ShieldTriangularWithRedInBlueFramePattern = "3846p47"
ShieldTriangularWithRed_PeachQuartersPattern = "3846p4t"
ShieldTriangularWithRoyalKnightsLionPattern = "3846p4d"
ShieldTriangularWithWolfpackPattern = "3846p44"
ShieldTriangularWithYellowLionOnBluePattern = "3846p4h"
ShieldTriangularWithYellowTrefoilsOnBlueSticker = "3846d04"
Shovel = "3837"
SignalHolder = "3900"
SignalHolderWithBlack_Police_AndRedLinePattern = "3900p01"
SignalHolderWithGreenCircleOnWhiteSticker = "3900d01"
Skate = "93555"
SkateboardWithBlackWheels_Complete_ = "42511c01"
SkateboardWithFourWheelClips = "42511"
SkateboardWithTwoWheelClips = "45917"
Skeleton_Shortcut_ = "6260c01"
Ski4LWithHinge = "6120"
Ski4LWithoutHinge = "99774"
Ski6L = "90509"
SkiPole = "90540"
Skirt = "36036"
Skirt0_7LWith11DiamondPoints = "16820"
Skirt0_7LWith11DiamondPoints_Formed_ = "16820c01"
Skirt1_1LWithStraightBottom = "600880"
Skirt1_1LWithStraightBottom_Formed_ = "600880c01"
Skirt1_1LWithStraightBottomWithDarkGreenApronPattern = "600880p01"
Skirt1_1LWithStraightBottomWithDarkGreenApronPattern_Formed_ = "600880p01c01"
Skirt1_4LWithStraightBottom = "u9209"
Skirt1_4LWithStraightBottom_Formed_ = "u9209c01"
Skirt1_4LWithStraightBottomWithWhiteApronWithLacePattern = "u9209p01"
Skirt1_4LWithStraightBottomWithWhiteApronWithLacePattern_Formed_ = "u9209p01c01"
Skirt1_5LFringed = "14295"
Skirt1_5LFringed_Formed_ = "14295c01"
Skirt1_5LFringedWithSteppedEdge = "18200"
Skirt1_5LFringedWithSteppedEdge_Formed_ = "18200c01"
Sledgehammer = "75904"
Slingshot = "20546"
SlingshotWithBrightLightYellowBandPattern = "20546p02"
SlingshotWithDarkTanBandPattern = "20546p01"
SmallBowWithArrow = "95051"
Snowboard2X6 = "18746"
Snowshoe = "30284"
SoccerBall = "72824"
SoccerBallWithBlackPentagonsPattern = "72824p01"
SpaceScannerTool = "30035"
Spear = "4497"
SpearTipFacetedWithBar0_4L = "27257"
SpearWithFourSideBlades = "43899"
Speargun = "30088"
StaffWithCrescentEnd = "95050"
StaffWithSphericalEnd = "95049"
Statuette = "90398"
SteakOnBone = "98372"
SteakOnBoneWithRedMeatPattern = "98372p01"
StretcherWithBottomHinges = "4714"
StretcherWithoutBottomHinges = "93140"
Suitcase_Closed_ = "4449-f1"
Suitcase_Open_ = "4449-f2"
SuitcaseWithIronManArmouredSuitMarkVStickers = "4449-f1d01"
SurfBoard2X6_5 = "90397"
SurfBoard2X6_5WithPinkFlamesPattern = "90397p02"
SurfBoard2X6_5WithWhiteShark_BlackOpenMouthAndGillsPattern = "90397p05"
SurfBoard2X10 = "6075"
SwordCutlass = "2530"
SwordDoubleBladeSerratedWithBarHolder = "13549"
SwordGreatsword = "59"
SwordKatana_DragonGuard_ = "93055"
SwordKatanaType1_OctagonalGuard_ = "30173a"
SwordKatanaType2_SquareGuard_ = "30173b"
SwordKatanaType3_SquareGuardWithSmoothEndcap_ = "21459"
SwordKhopesh = "93247"
SwordLongsword = "98370"
SwordRapier = "93550"
SwordRomanGladius = "95673"
SwordSaberWithClipPommel = "59229"
SwordScabbardWithShoulderStrap = "95348"
SwordScimitar = "43887"
SwordScimitarWithJaggedEdge = "60752"
SwordShortsword = "3847"
SwordShortswordWithElaborateHilt = "76764"
SwordSmallWithAngularGuard = "95053"
SwordSmallWithCurvedBlade = "10053"
SwordUruk_Hai = "10050"
SwordWithAngularHilt = "48495"
Syringe = "87989"
TeddyBearWithArmsDown = "98382"
TeddyBearWithArmsDownWithTanBellyAndMuzzleAndBlackNoseAndEyesPattern = "98382p01"
Telescope = "64644"
TennisRacket = "93216"
TomahawkWithFlat_SilverBlade = "13571"
Tool4_WayLugWrench = "604553"
ToolAdjustableWrenchWith3_RibHandle = "604614"
ToolBatteryPoweredDrill = "604549"
ToolBinocularsSpace = "30304"
ToolBoxWrench = "55300"
ToolBoxWrenchWith3_RibHandle = "604552"
ToolCrowbar = "92585"
ToolFishingRod = "2614"
ToolGrapplingHook = "30192"
ToolHammer = "55295"
ToolHammerWith3_RibHandle = "604547"
ToolHandaxe = "3835"
ToolHolder = "2516"
ToolHoseNozzleWithHandle = "4210a"
ToolMagnifyingGlass = "30152"
ToolMallet = "4522"
ToolOar = "2542"
ToolOarReinforced = "87585"
ToolOilcan = "55296"
ToolOilcanWithRibbedHandle = "604548"
ToolOpenEndWrench = "55299"
ToolOpenEndWrenchWith3_RibHandle = "604551"
ToolPowerDrill = "55297"
ToolScissors = "18920"
ToolScrewdriver = "55298"
ToolSocketWrenchWithRatchetAnd3_RibHandle = "604615"
ToolSpanner_Screwdriver = "4006"
ToolWand = "36752a"
ToolWandSprue = "36752b"
ToolWandsOnSprue = "36752c01"
Toolbox1X3WithHandle = "98368"
Torch = "3959"
TorchWithoutGrooves = "86208"
ToyWinderKey = "98375"
TrayOval = "11252"
TrophyCup2_4L = "10172"
TrophyCup3_6L = "89801"
TrophyCup5L = "72092"
Umbrella = "4094b"
UmbrellaFolded = "27150"
UnderwaterScooter = "30092"
UtilityBelt = "27145"
Vest = "3840"
VestWithCrownOnDarkPinkSticker = "3840d01"
VestWithCrownOnVioletSticker = "3840d05"
VestWithGreenChevronsOnYellowSticker = "3840d03"
VestWithGreenChevronsOnYellow_LtgreySticker = "3840d07"
VestWithWhiteMalteseCrossOnRedSticker = "3840d02"
VestWithWhiteMalteseCrossOnRed_LtgreySticker = "3840d06"
VestWithYellowTrefoilsOnBlueSticker = "3840d04"
VestWithYellowTrefoilsOnDkblueSticker = "3840d08"
WeaponBillyClub = "13790"
WeaponBladedClaw = "88811"
WeaponClubWithSpikes = "88001"
WeaponCrescentBladeSerratedWithBar0_5L = "98141"
WeaponCurvedBlade8_5LWithBar1_5L = "98137"
WeaponTrident = "92290"
WeldingGasTorchWithHoseAndGasCylinderTop = "13793"
Whip = "2488"
WhipBentFlexible = "88704"
WhipCoiled = "61975"
WhipInLatchedPosition = "2488c01"
WineGlass = "33061"
ZipLineHandle_NeedsWork_ = "30229"
